#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import warnings
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

import httpx

from argilla.client.sdk.commons.errors_handler import handle_response_error
from argilla.client.sdk.commons.models import ErrorMessage, HTTPValidationError, Response
from argilla.client.sdk.v1.datasets.models import (
    FeedbackDatasetModel,
    FeedbackFieldModel,
    FeedbackListVectorSettingsModel,
    FeedbackMetadataPropertyModel,
    FeedbackMetricsModel,
    FeedbackQuestionModel,
    FeedbackRecordsModel,
    FeedbackRecordsSearchModel,
    FeedbackRecordsSearchVectorQuery,
    FeedbackResponseStatusFilter,
    FeedbackVectorSettingsModel,
)


def create_dataset(
    client: httpx.Client,
    name: str,
    workspace_id: Union[str, UUID],
    guidelines: Optional[str] = None,
    allow_extra_metadata: bool = True,
) -> Response[Union[FeedbackDatasetModel, ErrorMessage, HTTPValidationError]]:
    """Sends a POST request to `/api/v1/datasets` endpoint to create a new `FeedbackDataset`.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        name: the name of the dataset to be created.
        workspace_id: the id of the workspace where the dataset will be created.
        guidelines: the guidelines of the dataset to be created. Defaults to `None`.
        allow_extra_metadata: whether to allow extra metadata not defined as a metadata property. Defaults to `True`.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if the
        request was successful, which is an instance of `FeedbackDatasetModel`.
    """
    url = "/api/v1/datasets"

    body = {"name": name, "workspace_id": str(workspace_id), "allow_extra_metadata": allow_extra_metadata}
    if guidelines is not None:
        body.update({"guidelines": guidelines})

    response = client.post(url=url, json=body)

    if response.status_code == 201:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = FeedbackDatasetModel(**response.json())
        return response_obj
    return handle_response_error(response)


def get_dataset(
    client: httpx.Client, id: UUID
) -> Response[Union[FeedbackDatasetModel, ErrorMessage, HTTPValidationError]]:
    """Sends a GET request to `/api/v1/datasets/{id}` endpoint to retrieve a `FeedbackDataset`.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the dataset to be retrieved.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if the
        request was successful, which is an instance of `FeedbackDatasetModel`.
    """
    url = f"/api/v1/datasets/{id}"

    response = client.get(url=url)

    if response.status_code == 200:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = FeedbackDatasetModel(**response.json())
        return response_obj
    return handle_response_error(response)


def delete_dataset(client: httpx.Client, id: UUID) -> Response[Union[ErrorMessage, HTTPValidationError]]:
    """Sends a DELETE request to `/api/v1/datasets/{id}` endpoint to delete a `FeedbackDataset`.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the dataset to be deleted.

    Returns:
        A `Response` object with the response itself, and/or the error codes if applicable.
    """
    url = f"/api/v1/datasets/{id}"

    response = client.delete(url=url)

    if response.status_code == 200:
        return Response.from_httpx_response(response)
    return handle_response_error(response)


def publish_dataset(
    client: httpx.Client, id: UUID
) -> Response[Union[FeedbackDatasetModel, ErrorMessage, HTTPValidationError]]:
    """Sends a PUT request to `/api/v1/datasets/{id}/publish` endpoint to publish a `FeedbackDataset`.
    Publishing in Argilla means setting the status of the dataset from `draft` to `ready`, so that
    it can be used to add records to it.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the dataset to be published.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if the
        request was successful, which is an instance of `FeedbackDatasetModel`.
    """
    url = f"/api/v1/datasets/{id}/publish"

    response = client.put(url=url)

    if response.status_code == 200:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = FeedbackDatasetModel(**response.json())
        return response_obj
    return handle_response_error(response)


def list_datasets(
    client: httpx.Client,
    workspace_id: Optional[UUID] = None,
) -> Response[Union[list, List[FeedbackDatasetModel], ErrorMessage, HTTPValidationError]]:
    """Sends a GET request to `/api/v1/me/datasets` endpoint to retrieve a list of
    `FeedbackTask` datasets filtered by `workspace_id` if applicable.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        workspace_id: the id of the workspace to filter the datasets by. Note that the user
            should either be owner or have access to the workspace. Defaults to None.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if the
        request was successful, which is a list of `FeedbackDatasetModel` if any, otherwise
        it will contain an empty list.
    """
    url = "/api/v1/me/datasets"

    params = {}
    if workspace_id is not None:
        params["workspace_id"] = str(workspace_id)

    response = client.get(url=url, params=params)

    if response.status_code == 200:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = [FeedbackDatasetModel(**dataset) for dataset in response.json()["items"]]
        return response_obj
    return handle_response_error(response)


def get_records(
    client: httpx.Client,
    id: UUID,
    include: Union[None, List[str]] = None,
    offset: int = 0,
    limit: int = 50,
    response_status: Optional[List[FeedbackResponseStatusFilter]] = None,
    metadata_filters: Optional[List[str]] = None,
    sort_by: Optional[List[str]] = None,
) -> Response[Union[FeedbackRecordsModel, ErrorMessage, HTTPValidationError]]:
    """Sends a GET request to `/api/v1/datasets/{id}/records` endpoint to retrieve a
    list of `FeedbackTask` records.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the dataset to retrieve the records from.
        include: the fields to be included in the response.
             Can either be `responses`, `suggestions`, `vectors:all` or `vectors:vector_name_1,vector_name_2,...`
        offset: the offset to be used in the pagination. Defaults to 0.
        limit: the limit to be used in the pagination. Defaults to 50.
        response_status: the status of the responses to be retrieved. Can either be
            `draft`, `missing`, `discarded`, or `submitted`. Defaults to None.
        metadata_filters: the metadata filters to be applied to the records. Defaults to None.
        sort_by: the fields to be used to sort the records. Defaults to None.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if the
        request was successful, which is an instance of `FeedbackRecordsModel`.
    """
    url = f"/api/v1/datasets/{id}/records"
    params = {"offset": offset, "limit": limit}

    if include:
        params["include"] = include

    if response_status:
        params["response_status"] = response_status

    if metadata_filters:
        params["metadata"] = metadata_filters

    if sort_by:
        params["sort_by"] = sort_by

    response = client.get(url=url, params=params)

    if response.status_code == 200:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = FeedbackRecordsModel(**response.json())
        return response_obj
    return handle_response_error(response)


def search_records(
    client: httpx.Client,
    id: UUID,
    vector_query: FeedbackRecordsSearchVectorQuery,
    include: Union[None, List[str]] = None,
    response_status: Optional[List[FeedbackResponseStatusFilter]] = None,
    metadata_filters: Optional[List[str]] = None,
    limit: int = 50,
) -> Response[Union[FeedbackRecordsSearchModel, ErrorMessage, HTTPValidationError]]:
    """Sends a POST request to `/api/v1/datasets/{id}/records/search` endpoint to search for records inside an specific dataset.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the dataset to add the records to.
        include: the fields to be included in the response.
        vector_query: the vector query to be used to search for records.
        response_status: the status of the responses to be retrieved.
            Can either be `draft`, `missing`, `discarded`, or `submitted`. Defaults to None.
        metadata_filters: the metadata filters to be applied to the records. Defaults to None.
        limit: an optional value to limit the number of returned records by the search.

    Returns:
        A `Response` object with the response itself, and/or the error codes if applicable.
    """
    url = f"/api/v1/datasets/{id}/records/search"

    params = {"limit": limit}
    if include:
        params["include"] = include
    if response_status:
        params["response_status"] = response_status
    if metadata_filters:
        params["metadata"] = metadata_filters

    vector_json = {"name": vector_query.name}
    if vector_query.value:
        vector_json["value"] = vector_query.value
    if vector_query.record_id:
        vector_json["record_id"] = str(vector_query.record_id)

    json = {"query": {"vector": vector_json}}

    response = client.post(url=url, params=params, json=json)

    if response.status_code == 200:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = FeedbackRecordsSearchModel(**response.json())

        return response_obj

    return handle_response_error(response)


def add_records(
    client: httpx.Client, id: UUID, records: List[Dict[str, Any]]
) -> Response[Union[ErrorMessage, HTTPValidationError]]:
    """Sends a POST request to `/api/v1/datasets/{id}/records` endpoint to add a list of `FeedbackTask` records.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the dataset to add the records to.
        records: the list of records to be added to the dataset.

    Returns:
        A `Response` object with the response itself, and/or the error codes if applicable.
    """
    url = f"/api/v1/datasets/{id}/records"

    active_user_id = None

    for record in records:
        cleaned_responses = []
        response_without_user_id = False
        for response in record.get("responses", []):
            if response.get("user_id") is None:
                if response_without_user_id:
                    warnings.warn(
                        f"Multiple responses without `user_id` found in record {record}, so just the first one will be"
                        " used while the rest will be ignored."
                    )
                    continue
                else:
                    if active_user_id is None:
                        active_user_id = client.get("api/me").json()["id"]
                    response["user_id"] = active_user_id
                response_without_user_id = True
            if isinstance(response.get("user_id"), UUID):
                response["user_id"] = str(response.get("user_id"))
            cleaned_responses.append(response)
        if len(cleaned_responses) > 0:
            record["responses"] = cleaned_responses

        for suggestion in record.get("suggestions", []):
            if isinstance(suggestion.get("question_id"), UUID):
                suggestion["question_id"] = str(suggestion.get("question_id"))

    response = client.post(url=url, json={"items": records})

    if response.status_code == 204:
        return Response.from_httpx_response(response)
    return handle_response_error(response)


def update_records(
    client: httpx.Client, id: UUID, records: List[Dict[str, Any]]
) -> Response[Union[ErrorMessage, HTTPValidationError]]:
    """Sends a PATCH requests to `/api/v1/datasets/{id}/records` endpoint to update a
    a list of `FeedbackTask` records from a `FeedbackDataset`.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the dataset to update the records from.
        records: the list of records to be updated.

    Returns:
        A `Response` object with the response itself, and/or the error codes if applicable.
    """
    url = f"/api/v1/datasets/{id}/records"

    items = []
    for record in records:
        item = {"id": record["id"]}
        if "metadata" in record:
            item["metadata"] = record["metadata"]
        if "suggestions" in record:
            item["suggestions"] = record["suggestions"]
        if "vectors" in record:
            item["vectors"] = record["vectors"]

        items.append(item)

    response = client.patch(url=url, json={"items": items})

    if response.status_code == 204:
        return Response.from_httpx_response(response)

    return handle_response_error(response)


def delete_records(
    client: httpx.Client, id: UUID, record_ids: List[UUID]
) -> Response[Union[ErrorMessage, HTTPValidationError]]:
    """Sends a DELETE request to `/api/v1/{id}/records` endpoint to remove a list of `FeedbackDataset` records.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the dataset to remove the records from.
        record_ids: the IDs of the records to be removed from the dataset.

    Returns:
        A `Response` object with the response itself, and/or the errors codes if applicable.
    """
    url = f"/api/v1/datasets/{id}/records"

    uuids_str = ",".join([str(record_id) for record_id in record_ids])
    response = client.delete(url=url, params={"ids": uuids_str})

    if response.status_code == 204:
        return Response.from_httpx_response(response)
    return handle_response_error(response)


def get_fields(
    client: httpx.Client, id: UUID
) -> Response[Union[List[FeedbackFieldModel], ErrorMessage, HTTPValidationError]]:
    """Sends a GET request to `/api/v1/datasets/{id}/fields` endpoint to retrieve a list of `FeedbackTask` fields.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the dataset to retrieve the fields from.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if the
        request was successful, which is a list of `FeedbackFieldModel`.
    """
    url = f"/api/v1/datasets/{id}/fields"

    response = client.get(url=url)

    if response.status_code == 200:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = [FeedbackFieldModel(**item) for item in response.json()["items"]]
        return response_obj
    return handle_response_error(response)


def add_field(
    client: httpx.Client, id: UUID, field: Dict[str, Any]
) -> Response[Union[FeedbackFieldModel, ErrorMessage, HTTPValidationError]]:
    """Sends a POST request to `/api/v1/datasets/{id}/fields` endpoint to add a `FeedbackTask` field.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the dataset to add the field to.
        field: the field to be added to the dataset.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if the
        request was successful, which is an instance of `FeedbackFieldModel`.
    """
    url = f"/api/v1/datasets/{id}/fields"

    response = client.post(url=url, json=field)

    if response.status_code == 201:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = FeedbackFieldModel(**response.json())
        return response_obj
    return handle_response_error(response)


def get_questions(
    client: httpx.Client, id: UUID
) -> Response[Union[List[FeedbackQuestionModel], ErrorMessage, HTTPValidationError]]:
    """Sends a GET request to `/api/v1/datasets/{id}/questions` endpoint to retrieve a
    list of `FeedbackTask` questions.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the dataset to retrieve the questions from.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if the
        request was successful, which is a list of `FeedbackQuestionModel`.
    """
    url = f"/api/v1/datasets/{id}/questions"

    response = client.get(url=url)

    if response.status_code == 200:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = [FeedbackQuestionModel(**item) for item in response.json()["items"]]
        return response_obj
    return handle_response_error(response)


def add_question(
    client: httpx.Client, id: UUID, question: Dict[str, Any]
) -> Response[Union[FeedbackQuestionModel, ErrorMessage, HTTPValidationError]]:
    """Sends a POST request to `/api/v1/datasets/{id}/questions` endpoint to add a
    question to the `FeedbackDataset`.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the dataset to add the question to.
        question: the question to be added to the dataset.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if the
        request was successful, which is a `FeedbackQuestionModel`.
    """
    url = f"/api/v1/datasets/{id}/questions"

    response = client.post(url=url, json=question)

    if response.status_code == 201:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = FeedbackQuestionModel(**response.json())
        return response_obj
    return handle_response_error(response)


def get_metadata_properties(
    client: httpx.Client, id: UUID
) -> Response[Union[List[FeedbackMetadataPropertyModel], ErrorMessage, HTTPValidationError]]:
    """Sends a GET request to `/api/v1/datasets/{id}/metadata-properties` endpoint to
    retrieve a list of `FeedbackDataset` metadata properties.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the dataset to retrieve the metadata properties from.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if the
        request was successful, which is a list of `FeedbackMetadataPropertyModel`.
    """
    url = f"/api/v1/me/datasets/{id}/metadata-properties"

    response = client.get(url=url)

    if response.status_code == 200:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = [FeedbackMetadataPropertyModel(**item) for item in response.json()["items"]]
        return response_obj
    return handle_response_error(response)


def add_metadata_property(
    client: httpx.Client, id: UUID, metadata_property: Dict[str, Any]
) -> Response[Union[FeedbackMetadataPropertyModel, ErrorMessage, HTTPValidationError]]:
    """Sends a POST request to `/api/v1/datasets/{id}/metadata-properties` endpoint to
    add a metadata property to the `FeedbackDataset`.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the dataset to add the metadata property to.
        metadata_property: the metadata property to be added to the dataset.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if the
        request was successful, which is a `FeedbackMetadataPropertyModel`.
    """
    url = f"/api/v1/datasets/{id}/metadata-properties"

    response = client.post(url=url, json=metadata_property)

    if response.status_code == 201:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = FeedbackMetadataPropertyModel(**response.json())
        return response_obj
    return handle_response_error(response)


def list_vectors_settings(
    client: httpx.Client,
    id: UUID,
) -> Response[Union[FeedbackListVectorSettingsModel, ErrorMessage, HTTPValidationError]]:
    """Sends a GET request to `/api/v1/datasets/{id}/vectors-settings` endpoint to
    retrieve the vectors settings of a `FeedbackDataset`.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the dataset to retrieve the vector settings from.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if the
        request was successful, which is a `FeedbackListVectorSettingsModel`.
    """
    url = f"/api/v1/datasets/{id}/vectors-settings"

    response = client.get(url=url)
    if response.status_code == 200:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = FeedbackListVectorSettingsModel(**response.json())
        return response_obj

    return handle_response_error(response)


def add_vector_settings(
    client: httpx.Client,
    id: UUID,
    name: str,
    title: str,
    dimensions: int,
) -> Response[Union[FeedbackVectorSettingsModel, ErrorMessage, HTTPValidationError]]:
    """Sends a POST request to `/api/v1/datasets/{id}/vectors-settings` endpoint to
    add a vector settings to the `FeedbackDataset`.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the dataset to add the vector settings to.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if
        the request was successful, which is a `FeedbackVectorSettingsModel`.
    """
    url = f"/api/v1/datasets/{id}/vectors-settings"

    body = {
        "name": name,
        "title": title,
        "dimensions": dimensions,
    }

    response = client.post(url=url, json=body)
    if response.status_code == 201:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = FeedbackVectorSettingsModel(**response.json())
        return response_obj

    # TODO: better handle error for v1 API endpoints
    return handle_response_error(response)


def get_metrics(
    client: httpx.Client,
    id: UUID,
) -> Response[Union[FeedbackMetricsModel, ErrorMessage, HTTPValidationError]]:
    """Sends a GET request to `/api/v1/datasets/{id}/metrics` endpoint to retrieve the metrics
    of a `FeedbackDataset`.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the dataset to retrieve the metrics from.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if the
        request was successful, which is a `FeedbackMetricsModel`.
    """
    url = f"/api/v1/me/datasets/{id}/metrics"

    response = client.get(url=url)

    if response.status_code == 200:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = FeedbackMetricsModel(**response.json())
        return response_obj
    return handle_response_error(response)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from argilla.client.feedback.schemas.enums import (
    FieldTypes,
    QuestionTypes,
    RecordSortField,
    ResponseStatusFilter,
    SortOrder,
)
from argilla.client.feedback.schemas.fields import FieldSchema, TextField
from argilla.client.feedback.schemas.metadata import (
    FloatMetadataFilter,
    FloatMetadataProperty,
    IntegerMetadataFilter,
    IntegerMetadataProperty,
    TermsMetadataFilter,
    TermsMetadataProperty,
)
from argilla.client.feedback.schemas.questions import (
    LabelQuestion,
    MultiLabelQuestion,
    QuestionSchema,
    RankingQuestion,
    RatingQuestion,
    TextQuestion,
)
from argilla.client.feedback.schemas.records import (
    FeedbackRecord,
    RankingValueSchema,
    ResponseSchema,
    SortBy,
    SuggestionSchema,
    ValueSchema,
)
from argilla.client.feedback.schemas.vector_settings import VectorSettings

__all__ = [
    "FieldTypes",
    "QuestionTypes",
    "FieldSchema",
    "TextField",
    "FloatMetadataProperty",
    "IntegerMetadataProperty",
    "TermsMetadataProperty",
    "TermsMetadataFilter",
    "IntegerMetadataFilter",
    "FloatMetadataFilter",
    "LabelQuestion",
    "MultiLabelQuestion",
    "QuestionSchema",
    "RankingQuestion",
    "RatingQuestion",
    "TextQuestion",
    "FeedbackRecord",
    "RankingValueSchema",
    "ResponseSchema",
    "SuggestionSchema",
    "ValueSchema",
    "SortOrder",
    "SortBy",
    "RecordSortField",
    "ResponseStatusFilter",
    "VectorSettings",
]



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import typer


def list_workspaces() -> None:
    """List the workspaces in Argilla and prints them on the console."""
    from rich.console import Console

    from argilla import Workspace
    from argilla.cli.rich import get_argilla_themed_table

    workspaces = Workspace.list()

    table = get_argilla_themed_table(title="Workspaces")
    for column in ("ID", "Name", "Creation Date", "Last Update Date"):
        table.add_column(column, justify="center")

    for workspace in workspaces:
        table.add_row(
            str(workspace.id),
            workspace.name,
            workspace.inserted_at.isoformat(sep=" "),
            workspace.updated_at.isoformat(sep=" "),
        )

    console = Console()
    console.print(table)


if __name__ == "__main__":
    typer.run(list_workspaces)



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Optional

import typer

from argilla.cli.callback import init_callback

from .add_user import add_user
from .create import create_workspace
from .delete_user import delete_user
from .list import list_workspaces

_COMMANDS_REQUIRING_WORKSPACE = ["add-user", "delete-user"]


def callback(
    ctx: typer.Context,
    name: Optional[str] = typer.Option(None, help="Name of the workspace to which apply the command."),
) -> None:
    init_callback()

    if ctx.invoked_subcommand not in _COMMANDS_REQUIRING_WORKSPACE:
        return

    if name is None:
        raise typer.BadParameter("The command requires a workspace name provided using '--name' option")

    from argilla.cli.rich import echo_in_panel
    from argilla.client.workspaces import Workspace

    try:
        workspace = Workspace.from_name(name)
    except ValueError as e:
        echo_in_panel(
            f"Workspace with name={name} does not exist.",
            title="Workspace not found",
            title_align="left",
            success=False,
        )
        raise typer.Exit(code=1) from e
    except RuntimeError as e:
        echo_in_panel(
            "An unexpected error occurred when trying to get the workspace from the Argilla server",
            title="Unexpected error",
            title_align="left",
            success=False,
        )
        raise typer.Exit(code=1) from e

    ctx.obj = workspace


app = typer.Typer(help="Commands for workspace management", no_args_is_help=True, callback=callback)

app.command(name="create", help="Create a workspace")(create_workspace)
app.command(name="list", help="Lists workspaces of the logged user.")(list_workspaces)
app.command(name="delete-user", help="Deletes a user from a workspace.")(delete_user)
app.command(name="add-user", help="Adds a user to a workspace.")(add_user)


if __name__ == "__main__":
    app()



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import List, Union

from datasets import DatasetDict

from argilla.client.models import TokenClassificationRecord
from argilla.training.base import ArgillaTrainerSkeleton
from argilla.training.utils import filter_allowed_args, get_default_args
from argilla.utils.dependency import require_dependencies


class ArgillaSpanMarkerTrainer(ArgillaTrainerSkeleton):
    _logger = logging.getLogger("ArgillaSpanMarkerTrainer")
    _logger.setLevel(logging.INFO)

    def __init__(self, *args, **kwargs) -> None:
        require_dependencies(["datasets", "span_marker>=1.2", "transformers>=4.19.0"])
        super().__init__(*args, **kwargs)

        import torch
        from span_marker import SpanMarkerModel

        self.trainer_model = None

        self.device = "cpu"
        if torch.backends.mps.is_available():
            self.device = "mps"
        elif torch.cuda.is_available():
            self.device = "cuda"

        if self._seed is None:
            self._seed = 42

        if self._model is None:
            self._model = "bert-base-cased"
            self._logger.warning(f"No model defined. Using the default model {self._model}.")

        if isinstance(self._dataset, DatasetDict):
            self._train_dataset = self._dataset["train"]
            self._eval_dataset = self._dataset["test"]
        else:
            self._train_dataset = self._dataset
            self._eval_dataset = None

        if self._record_class == TokenClassificationRecord:
            self._column_mapping = {"text": "text", "token": "tokens", "ner_tags": "ner_tags"}
            self._label_list = self._train_dataset.features["ner_tags"].feature.names

            self._model_class = SpanMarkerModel
        else:
            raise NotImplementedError("Text2TextRecord and TextClassification are not supported.")

        self.init_training_args()

    def init_training_args(self) -> None:
        from transformers import TrainingArguments

        self.model_kwargs = {"pretrained_model_name_or_path": self._model, "labels": self._label_list}

        self.trainer_kwargs = get_default_args(TrainingArguments.__init__)
        self.trainer_kwargs["evaluation_strategy"] = "no" if self._eval_dataset is None else "epoch"
        self.trainer_kwargs["logging_steps"] = 30
        self.trainer_kwargs["learning_rate"] = 5e-5
        self.trainer_kwargs["weight_decay"] = 0.01

    def init_model(self) -> None:
        self.trainer_model = self._model_class.from_pretrained(**self.model_kwargs).to(self.device)

    def update_config(self, **kwargs) -> None:
        """
        Updates the `model_kwargs` and `trainer_kwargs` dictionaries with the keyword
        arguments passed to the `update_config` function.
        """
        from span_marker import SpanMarkerConfig
        from transformers import TrainingArguments

        self.model_kwargs.update(filter_allowed_args(SpanMarkerConfig.__init__, **kwargs))
        self.trainer_kwargs.update(filter_allowed_args(TrainingArguments.__init__, **kwargs))

    def __repr__(self):
        formatted_string = []
        arg_dict = {"'SpanMarkerModel'": self.model_kwargs, "'Trainer'": self.trainer_kwargs}
        for arg_dict_key, arg_dict_single in arg_dict.items():
            formatted_string.append(arg_dict_key)
            for key, val in arg_dict_single.items():
                formatted_string.append(f"{key}: {val}")
        return "\n".join(formatted_string)

    def train(self, output_dir: str):
        """
        We create a SetFitModel object from a pretrained model, then create a SetFitTrainer object with
        the model, and then train the model
        """
        from span_marker import Trainer
        from transformers import TrainingArguments

        self.trainer_kwargs["output_dir"] = output_dir

        # prepare data
        self.init_model()

        # set trainer
        self.trainer = Trainer(
            args=TrainingArguments(**self.trainer_kwargs),
            model=self.trainer_model,
            train_dataset=self._train_dataset,
            eval_dataset=self._eval_dataset,
        )

        # train
        self.trainer.train()
        if self._eval_dataset:
            self._metrics = self.trainer.evaluate()
            self._logger.info(self._metrics)
        else:
            self._metrics = None

        self.save(output_dir)

    def predict(self, text: Union[List[str], str], as_argilla_records: bool = True, **kwargs):
        """
        The function takes in a list of strings and returns a list of predictions

        Args:
          text (Union[List[str], str]): The text to be classified.
          as_argilla_records (bool): If True, the prediction will be returned as an Argilla record. If
        False, the prediction will be returned as a string. Defaults to True

        Returns:
          A list of predictions
        """
        from datasets import Dataset

        if self.trainer_model is None:
            self._logger.warning("Using model without fine-tuning.")
            self.init_model()

        str_input = False
        if isinstance(text, str):
            text = [text]
            str_input = True

        entities_list = self.trainer_model.predict(text, **kwargs)

        if as_argilla_records:
            formatted_prediction = []

            for sentence, entities in zip(text, entities_list):
                predictions = [
                    (entity["label"], entity["char_start_index"], entity["char_end_index"], entity["score"])
                    for entity in entities
                ]
                dataset = Dataset.from_dict({"tokens": text})
                encoding = self.trainer_model.tokenizer({"tokens": dataset["tokens"]}, return_batch_encoding=True)[
                    "batch_encoding"
                ]
                word_ids = sorted(set(encoding.word_ids()) - {None})
                tokens = []
                for word_id in word_ids:
                    char_span = encoding.word_to_chars(word_id)
                    tokens.append(sentence[char_span.start : char_span.end].lstrip().rstrip())
                formatted_prediction.append(self._record_class(text=sentence, tokens=tokens, prediction=predictions))
        else:
            formatted_prediction = entities_list

        if str_input:
            return formatted_prediction[0]
        return formatted_prediction

    def save(self, output_dir: str):
        """
        The function saves the model to the path specified, and also saves the label2id and id2label
        dictionaries to the same path

        Args:
          output_dir (str): the path to save the model to
        """
        self.trainer_model.save_pretrained(output_dir)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import typer

from argilla.client.models import Framework

app = typer.Typer(no_args_is_help=True)


def framework_callback(value: str):
    try:
        return Framework(value.lower())
    except ValueError:
        raise typer.BadParameter(f"Invalid framework {value}. Choose from {', '.join([f.value for f in Framework])}")


# using callback to ensure it is used as sole command
@app.callback(help="Starts the ArgillaTrainer", invoke_without_command=True)
def train(
    name: str = typer.Option(default=None, help="The name of the dataset to be used for training."),
    framework: Framework = typer.Option(default=None, help="The framework to be used for training."),
    workspace: str = typer.Option(default=None, help="The workspace to be used for training."),
    limit: int = typer.Option(default=None, help="The number of record to be used."),
    query: str = typer.Option(default=None, help="The query to be used."),
    model: str = typer.Option(default=None, help="The modelname or path to be used for training."),
    train_size: float = typer.Option(default=1.0, help="The train split to be used."),
    seed: int = typer.Option(default=42, help="The random seed number."),
    device: int = typer.Option(default=-1, help="The GPU id to be used for training."),
    output_dir: str = typer.Option(default="model", help="Output directory for the saved model."),
    update_config_kwargs: str = typer.Option(default={}, help="update_config() kwargs to be passed as a dictionary."),
):
    import json

    from argilla.cli.callback import init_callback
    from argilla.client.singleton import init
    from argilla.training import ArgillaTrainer

    init_callback()

    trainer = ArgillaTrainer(
        name=name,
        framework=framework,
        workspace=workspace,
        model=model,
        train_size=train_size,
        seed=seed,
        gpu_id=device,
        limit=limit,
        query=query,
    )
    trainer.update_config(**json.loads(update_config_kwargs))
    if output_dir:
        trainer.train(output_dir=output_dir)
    else:
        trainer.train()


if __name__ == "__main__":
    app()



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import functools
import warnings
from dataclasses import dataclass
from typing import Callable, List, TypeVar

try:
    from typing import ParamSpec
except ImportError:
    from typing_extensions import ParamSpec

from argilla.client.apis.status import Status
from argilla.client.sdk.users import api as users_api
from argilla.client.sdk.users.models import UserRole
from argilla.client.singleton import active_client

_P = ParamSpec("_P")
_R = TypeVar("_R")


def allowed_for_roles(roles: List[UserRole]) -> Callable[[Callable[_P, _R]], Callable[_P, _R]]:
    """Decorator function to check the role of the user calling the function, to restrict
    access to certain Argilla functions for users with certain roles. Note that the method
    inside the class to be decorated must contain the `_client` attribute with the active
    client at the class initialization time; otherwise, the current active client will be
    used instead, which in most of the cases won't match the client used during the class
    initialization.

    Args:
        roles: List of roles that are allowed to call the function.

    Example:
        >>> @allowed_for_roles(["owner"])
        >>> def delete_user(user_id: UUID) -> None:
        >>>     ...
    """

    def decorator(func: Callable[_P, _R]) -> Callable[_P, _R]:
        @functools.wraps(func)
        def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _R:
            client = args[0]._client if hasattr(args[0], "_client") else active_client().http_client.httpx
            user = users_api.whoami_httpx(client).parsed
            if user.role not in roles:
                raise PermissionError(
                    f"User with role={user.role} is not allowed to call `{func.__name__}`."
                    f" Only users with role={roles} are allowed to call this function."
                )
            return func(*args, **kwargs)

        return wrapper

    return decorator


@dataclass
class ServerInfo:
    url: str
    version: str
    elasticsearch_version: str


def server_info() -> ServerInfo:
    """Returns the information about the Argilla server that is currently active.

    Note that the connections to Argilla are established via `rg.init`, `argilla login` or
    setting the `ARGILLA_API_URL` and `ARGILLA_API_KEY` env variables.

    Returns:
        A `dataclass` with the following fields: `url`, `version` and `elasticsearch_version`.
    """
    # Filtering the warnings to avoid some redundant and unrelated warnings
    warnings.filterwarnings("ignore", category=UserWarning)

    try:
        client = active_client().client
        info = Status(client).get_status()
    except Exception as e:
        raise RuntimeError(
            "You must be logged in to Argilla to use this function."
            " Please call `rg.init`, run `argilla login` or set the `ARGILLA_API_URL` and `ARGILLA_API_KEY` env variables."
            " If you already did it, please check that the Argilla server is up and running."
        ) from e

    return ServerInfo(
        url=client.base_url,
        version=info.version,
        elasticsearch_version=(
            f"{info.elasticsearch.version.number} ({info.elasticsearch.version.distribution})"
            if info.elasticsearch.version.distribution
            else info.elasticsearch.version.number
        ),
    )



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import asyncio
import importlib
import os
import sys
import threading
import warnings
from itertools import chain
from types import ModuleType
from typing import Any, Optional, Tuple


class LazyargillaModule(ModuleType):
    """Module class that surfaces all objects but only performs associated imports when the objects are requested.

    Shamelessly copied and adapted from the Hugging Face transformers implementation.
    """

    def __init__(
        self,
        name,
        module_file,
        import_structure,
        deprecated_import_structure=None,
        module_spec=None,
        extra_objects=None,
    ):
        super().__init__(name)
        self._modules = set(import_structure.keys())
        self._class_to_module = {}
        for key, values in import_structure.items():
            for value in values:
                self._class_to_module[value] = key
        # Needed for autocompletion in an IDE
        self.__all__ = list(import_structure.keys()) + list(chain(*import_structure.values()))
        self.__file__ = module_file
        self.__spec__ = module_spec
        self.__path__ = [os.path.dirname(module_file)]
        self._objects = {} if extra_objects is None else extra_objects
        self._name = name
        self._import_structure = import_structure

        # deprecated stuff
        deprecated_import_structure = deprecated_import_structure or {}
        self._deprecated_modules = set(deprecated_import_structure.keys())
        self._deprecated_class_to_module = {}
        for key, values in deprecated_import_structure.items():
            for value in values:
                self._deprecated_class_to_module[value] = key

    # Needed for autocompletion in an IDE
    def __dir__(self):
        result = super().__dir__()
        # The elements of self.__all__ that are submodules may or may not be in the dir already, depending on whether
        # they have been accessed or not. So we only add the elements of self.__all__ that are not already in the dir.
        for attr in self.__all__:
            if attr not in result:
                result.append(attr)
        return result

    def __getattr__(self, name: str) -> Any:
        if name in self._objects:
            return self._objects[name]
        if name in self._modules:
            value = self._get_module(name)
        elif name in self._class_to_module.keys():
            module = self._get_module(self._class_to_module[name])
            value = getattr(module, name)
        elif name in self._deprecated_modules:
            value = self._get_module(name, deprecated=True)
        elif name in self._deprecated_class_to_module.keys():
            module = self._get_module(self._deprecated_class_to_module[name], deprecated=True, class_name=name)
            value = getattr(module, name)
        else:
            raise AttributeError(f"module {self.__name__} has no attribute {name}")

        setattr(self, name, value)
        return value

    def _get_module(
        self,
        module_name: str,
        deprecated: bool = False,
        class_name: Optional[str] = None,
    ):
        if deprecated:
            warnings.warn(
                f"Importing '{class_name or module_name}' from the argilla namespace (that is "
                f"`argilla.{class_name or module_name}`) is deprecated and will not work in a future version. "
                f"Make sure you update your code accordingly.",
                category=FutureWarning,
            )

        try:
            return importlib.import_module("." + module_name, self.__name__)
        except Exception as e:
            raise RuntimeError(
                f"Failed to import {self.__name__}.{module_name} because of the following error "
                f"(look up to see its traceback):\n{e}"
            ) from e

    def __reduce__(self):
        return self.__class__, (self._name, self.__file__, self._import_structure)


def limit_value_length(data: Any, max_length: int) -> Any:
    """
    Given an input data, limits string values to a max_length by fetching
    last max_length characters

    Parameters
    ----------
    data:
        Input data
    max_length:
        Max length for string values

    Returns
    -------
        Limited version of data, if any
    """

    if isinstance(data, str):
        return data[-max_length:]
    if isinstance(data, dict):
        return {k: limit_value_length(v, max_length=max_length) for k, v in data.items()}
    if isinstance(data, (list, tuple, set)):
        new_values = map(lambda x: limit_value_length(x, max_length=max_length), data)
        return type(data)(new_values)
    return data


__LOOP__, __THREAD__ = None, None


def setup_loop_in_thread() -> Tuple[asyncio.AbstractEventLoop, threading.Thread]:
    """Sets up a new asyncio event loop in a new thread, and runs it forever.

    Returns:
        A tuple containing the event loop and the thread.
    """

    def start_background_loop(loop: asyncio.AbstractEventLoop) -> None:
        asyncio.set_event_loop(loop)
        loop.run_forever()

    global __LOOP__
    global __THREAD__

    if not (__LOOP__ and __THREAD__):
        loop = asyncio.new_event_loop()
        thread = threading.Thread(target=start_background_loop, args=(loop,), daemon=True)
        thread.start()
        __LOOP__, __THREAD__ = loop, thread
    return __LOOP__, __THREAD__



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Optional

import typer


def push_to_huggingface(
    ctx: typer.Context,
    repo_id: str = typer.Option(..., help="The HuggingFace Hub repo where the dataset will be pushed to"),
    private: bool = typer.Option(False, help="Whether the dataset should be private or not"),
    token: Optional[str] = typer.Option(None, help="The HuggingFace Hub token to be used for pushing the dataset"),
) -> None:
    """Pushes a dataset from Argilla into the HuggingFace Hub. Note that this command
    is just available for `FeedbackDataset` datasets."""
    from rich.live import Live
    from rich.spinner import Spinner

    from argilla.cli.rich import echo_in_panel
    from argilla.client.feedback.dataset.local.dataset import FeedbackDataset

    dataset: "FeedbackDataset" = ctx.obj

    spinner = Spinner(
        name="dots",
        text=f"Pushing `FeedbackDataset` with name={dataset.name} and workspace={dataset.workspace.name} to the"
        " HuggingFace Hub...",
        style="red",
    )

    try:
        with Live(spinner, refresh_per_second=20):
            dataset.push_to_huggingface(repo_id=repo_id, private=private, token=token)
    except ValueError as e:
        echo_in_panel(
            "The `FeedbackDataset` has no records to push to the HuggingFace Hub. Make sure to add records before"
            " pushing it.",
            title="No records to push",
            title_align="left",
            success=False,
        )
        raise typer.Exit(1) from e
    except Exception as e:
        echo_in_panel(
            "An unexpected error occurred when trying to push the `FeedbackDataset` to the HuggingFace Hub",
            title="Unexpected error",
            title_align="left",
            success=False,
        )
        raise typer.Exit(code=1) from e

    echo_in_panel(
        f"`FeedbackDataset` successfully pushed to the HuggingFace Hub at https://huggingface.co/{repo_id}",
        title="Dataset pushed",
        title_align="left",
    )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import TYPE_CHECKING

from argilla.client.feedback.training.frameworks.transformers import ArgillaTransformersTrainer
from argilla.client.models import TextClassificationRecord
from argilla.training.setfit import ArgillaSetFitTrainer as ArgillaSetFitTrainerV1
from argilla.utils.dependency import require_dependencies, requires_dependencies

if TYPE_CHECKING:
    from argilla.client.feedback.integrations.huggingface.model_card import SetFitModelCardData


class ArgillaSetFitTrainer(ArgillaSetFitTrainerV1, ArgillaTransformersTrainer):
    _logger = logging.getLogger("ArgillaSetFitTrainer")
    _logger.setLevel(logging.INFO)

    require_dependencies(["torch", "datasets", "transformers", "setfit>=0.6"])

    def __init__(self, *args, **kwargs):
        if kwargs.get("model") is None and "model" in kwargs:
            kwargs["model"] = "all-MiniLM-L6-v2"
        self.multi_target_strategy = None
        self._column_mapping = None
        ArgillaTransformersTrainer.__init__(self, *args, **kwargs)

        if self._record_class is not TextClassificationRecord:
            raise NotImplementedError("SetFit only supports the `TextClassification` task.")

        if self._multi_label:
            self._column_mapping = {"text": "text", "binarized_label": "label"}
            self.multi_target_strategy = "one-vs-rest"
        else:
            self.multi_target_strategy = None
            self._column_mapping = {"text": "text", "label": "label"}
        self.init_training_args()

    def get_model_card_data(self, **card_data_kwargs) -> "SetFitModelCardData":
        """
        Generate the card data to be used for the `ArgillaModelCard`.

        Args:
            card_data_kwargs: Extra arguments provided by the user when creating the `ArgillaTrainer`.

        Returns:
            SetFitModelCardData: Container for the data to be written on the `ArgillaModelCard`.
        """
        from argilla.client.feedback.integrations.huggingface.model_card import SetFitModelCardData

        return SetFitModelCardData(
            model_id=self._model,
            task=self._task,
            update_config_kwargs={**self.model_kwargs, **self.trainer_kwargs},
            **card_data_kwargs,
        )

    @requires_dependencies("huggingface_hub")
    def push_to_huggingface(self, repo_id: str, **kwargs) -> None:
        """Uploads the model to [huggingface's model hub](https://huggingface.co/models).

        The full list of parameters can be seen at:
        [huggingface_hub](https://huggingface.co/docs/huggingface_hub/package_reference/mixins#huggingface_hub.ModelHubMixin.push_to_hub).

        Args:
            repo_id:
                The name of the repository you want to push your model and tokenizer to.
                It should contain your organization name when pushing to a given organization.

        Raises:
            NotImplementedError: If the model doesn't exist, meaning it hasn't been instantiated yet.
        """
        if not self._trainer:
            raise ValueError("The `trainer` must be initialized prior to this point. You should call `train`.")
        url = self._trainer.push_to_hub(repo_id, **kwargs)
        self._logger.info(f"Model pushed to: {url}")



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING, Iterator, List, Union

from argilla.client.feedback.constants import FETCHING_BATCH_SIZE
from argilla.client.feedback.schemas.remote.records import RemoteFeedbackRecord
from argilla.client.sdk.users.models import UserRole
from argilla.client.utils import allowed_for_roles

if TYPE_CHECKING:
    from argilla.client.feedback.dataset.remote.dataset import RemoteFeedbackRecords


class ArgillaRecordsMixin:
    @allowed_for_roles(roles=[UserRole.owner, UserRole.admin])
    def __getitem__(
        self: "RemoteFeedbackRecords", key: Union[slice, int]
    ) -> Union["RemoteFeedbackRecord", List["RemoteFeedbackRecord"]]:
        """Returns the record(s) at the given index(es) from Argilla.

        Args:
            key: the index(es) of the record(s) to return. Can either be a single index or a slice.

        Returns:
            Either the record of the given index, or a list with the records at the given indexes.
        """
        try:
            num_records = len(self)
        except NotImplementedError:
            num_records = None
            if not isinstance(key, int):
                raise NotImplementedError(f"`key`={key} is not supported for this dataset. Only `int` is supported.")

        offsets = []
        limit = None

        if isinstance(key, slice) and num_records is not None:
            start, stop, step = key.indices(num_records)
            if step is not None and step != 1:
                raise ValueError("When providing a `slice` just `step=None` or `step=1` are allowed.")
            if start < 0:
                start += num_records
            if stop < 0:
                stop += num_records
            if start < 0 or stop < 0:
                raise IndexError(
                    f"Index {start if start < 0 else stop} is out of range, dataset has {num_records} records."
                )
            limit = stop - start
            if limit < 0:
                raise ValueError("Negative slice bounds are not supported.")
            elif limit < FETCHING_BATCH_SIZE:
                offsets = [start]
                limits = [limit]
            else:
                offsets = list(range(start, stop, FETCHING_BATCH_SIZE))
                limits = [FETCHING_BATCH_SIZE] * len(offsets)
                if stop % FETCHING_BATCH_SIZE != 0:
                    offsets[-1] = stop - (stop % FETCHING_BATCH_SIZE) + 1
                    limits[-1] = (stop % FETCHING_BATCH_SIZE) - 1
        elif isinstance(key, int):
            if num_records is not None:
                if key < 0:
                    key += num_records
                if key < 0 or key >= num_records:
                    raise IndexError(f"Index {key} is out of range, dataset has {num_records} records.")
            offsets = [key]
            limits = [1]
        else:
            raise TypeError("Only `int` and `slice` are supported as index.")

        question_id_to_name = {question.id: question.name for question in self.dataset.questions}

        records = []
        for offset, limit in zip(offsets, limits):
            fetched_records = self._fetch_records(offset=offset, limit=limit)
            if len(fetched_records.items) == 0:
                break
            records.extend(
                [
                    RemoteFeedbackRecord.from_api(record, question_id_to_name=question_id_to_name, client=self._client)
                    for record in fetched_records.items
                ]
            )
        if len(records) == 0:
            raise IndexError(
                "No records were found in the dataset in Argilla for the given index(es) and/or filter(s) if any."
            )
        return records[0] if isinstance(key, int) else records

    @allowed_for_roles(roles=[UserRole.owner, UserRole.admin])
    def __iter__(self: "RemoteFeedbackRecords") -> Iterator["RemoteFeedbackRecord"]:
        """Iterates over the `FeedbackRecord`s of the current `FeedbackDataset` in Argilla."""
        current_batch = 0
        while True:
            batch = self._fetch_records(offset=FETCHING_BATCH_SIZE * current_batch, limit=FETCHING_BATCH_SIZE)
            if len(batch.items) == 0:
                break
            for record in batch.items:
                yield RemoteFeedbackRecord.from_api(
                    record, question_id_to_name=self._question_id_to_name, client=self._client
                )
            current_batch += 1

            if len(batch.items) < FETCHING_BATCH_SIZE:
                break



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING

import typer

if TYPE_CHECKING:
    from argilla.client.feedback.dataset.remote.dataset import RemoteFeedbackDataset


def delete_dataset(ctx: typer.Context) -> None:
    from argilla.cli.rich import echo_in_panel

    dataset: "RemoteFeedbackDataset" = ctx.obj

    try:
        dataset.delete()
    except RuntimeError as e:
        echo_in_panel(
            "An unexpected error occurred when trying to delete the `FeedbackDataset`",
            title="Unexpected error",
            title_align="left",
            success=False,
        )
        raise typer.Exit(code=1) from e

    echo_in_panel(
        f"`FeedbackDataset` with name={dataset.name} and workspace={dataset.workspace.name} deleted successfully",
        title="Dataset deleted",
        title_align="left",
    )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""This module contains metrics for Suggestions Metric and Responses Metric."""

import random
import warnings
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from argilla.client.feedback.dataset import FeedbackDataset
from argilla.client.feedback.metrics.base import AnnotatorMetricBase, MetricBase, ModelMetricResult
from argilla.client.feedback.metrics.utils import (
    get_responses_and_suggestions_per_user,
    get_unified_responses_and_suggestions,
    is_multiclass,
    map_str_to_int,
)
from argilla.client.feedback.schemas import (
    LabelQuestion,
    MultiLabelQuestion,
    RankingQuestion,
    RatingQuestion,
    TextQuestion,
)
from argilla.client.feedback.schemas.enums import ResponseStatusFilter
from argilla.client.feedback.schemas.records import SortBy
from argilla.utils.dependency import requires_dependencies

if TYPE_CHECKING:
    from argilla.client.feedback.dataset import FeedbackDataset
    from argilla.client.feedback.metrics.base import Responses, Suggestions
    from argilla.client.feedback.schemas.enums import ResponseStatusFilter
    from argilla.client.feedback.schemas.records import SortBy


class AnnotatorMetric(MetricBase):
    """Main class to compute annotator metrics. Annotator metrics refers to the combination of Suggestions Metric and Responses Metric. They are both different from the Agreement Metric (i.e. Inter-Annotator Agreement) and they are utilized to compute metrics contrasting suggestions vs responses.

    Example:
        >>> import argilla as rg
        >>> from argilla.client.feedback.metrics import AnnotatorMetric
        >>> metric = AnnotatorMetric(dataset=dataset, question_name=question)
        >>> metrics_report = metric.compute("accuracy")

    """

    def __init__(
        self,
        dataset: "FeedbackDataset",
        question_name: str,
        filter_by: Optional[Dict[str, Union["ResponseStatusFilter", List["ResponseStatusFilter"]]]] = None,
        sort_by: Optional[List["SortBy"]] = None,
        max_records: Optional[int] = None,
        responses_vs_suggestions: bool = True,
    ) -> None:
        """Initialize an `AnnotatorMetric` object to compute agreement metrics for both Suggestions Metric and Responses Metric.

        Args:
            dataset: FeedbackDataset to compute the metrics.
            question_name: Name of the question for which we want to analyse the agreement.
            filter_by: A dict with key the field to filter by, and values the filters to apply.
                Can be one of: draft, pending, submitted, and discarded. If set to None,
                no filter will be applied. Defaults to None (no filter is applied).
            sort_by: A list of `SortBy` objects to sort your dataset by.
                Defaults to None (no filter is applied).
            max_records: The maximum number of records to use for training. Defaults to None.
            responses_vs_suggestions: Whether to utilize Suggestions Metric (where the suggestions are the ground truths and the responses are compared against them) or Responses Metric (where the responses are the ground truths and the suggestions are compared against them). Defaults to True, i.e. Responses Metric.
        """
        self._metrics_per_question = METRICS_PER_QUESTION
        super().__init__(dataset, question_name, responses_vs_suggestions=responses_vs_suggestions)
        self._filter_by = filter_by
        self._sort_by = sort_by
        self._max_records = max_records

    def _check_responses_and_suggestions(
        self, responses_per_user: Dict[int, "Responses"], suggestions: "Suggestions"
    ) -> Tuple[Dict[int, "Responses"], "Suggestions"]:
        # Check for possible missing suggestions
        df_suggestions = pd.Series(suggestions)
        df_responses_per_user = pd.DataFrame(responses_per_user)
        df_responses_per_user = df_responses_per_user[df_suggestions.notna()]
        df_suggestions = df_suggestions[df_suggestions.notna()]
        total_responses = len(suggestions)

        responses_per_user = df_responses_per_user.to_dict(orient="list")
        suggestions = df_suggestions.to_list()

        if len(suggestions) == 0:
            raise ValueError("All the suggestions are None, the metric cannot be computed.")
        elif len(suggestions) < total_responses:
            warnings.warn("Some suggestions are None, the metric will be computed without them.")
        return responses_per_user, suggestions

    def compute(
        self, metric_names: Union[str, List[str]], show_progress: bool = True
    ) -> Dict[str, List[ModelMetricResult]]:
        """Computes the annotator metrics for the given question.

        Args:
            metric_names: name or list of names for the metrics to compute. i.e. `accuracy`

        Raises:
            ValueError: If the metric name is not supported for the given question.

        Returns:
            metrics: dict with the metrics computed for each annotator, where the
                key corresponds to the user id and the values are a list with the
                metric results.
        """
        metric_names = self._check_metrics(metric_names)
        metric_classes = self._get_metric_classes(metric_names)

        responses_and_suggestions_per_user = get_responses_and_suggestions_per_user(
            self._dataset,
            self._question_name,
            filter_by=self._filter_by,
            sort_by=self._sort_by,
            max_records=self._max_records,
        )
        metrics = defaultdict(list)
        for user_id, resp_and_suggest in responses_and_suggestions_per_user.items():
            responses = resp_and_suggest["responses"]
            suggestions = resp_and_suggest["suggestions"]
            as_responses, as_suggestions = self._prepare_responses_and_suggestions(responses, suggestions)
            for metric_name, metric_cls in metric_classes:
                metric = metric_cls(responses=as_responses, suggestions=as_suggestions)
                result = metric.compute()
                metrics[user_id].append(ModelMetricResult(metric_name=metric_name, result=result, count=len(responses)))

        return dict(metrics)


class ModelMetric(AnnotatorMetric):
    """Where suggestions are the ground truths and the responses are compared against them."""

    def __init__(
        self,
        dataset: FeedbackDataset,
        question_name: str,
        filter_by: Optional[Dict[str, Union["ResponseStatusFilter", List["ResponseStatusFilter"]]]] = None,
        sort_by: Optional[List["SortBy"]] = None,
        max_records: Optional[int] = None,
    ) -> None:
        super().__init__(
            dataset,
            question_name,
            filter_by=filter_by,
            sort_by=sort_by,
            max_records=max_records,
            responses_vs_suggestions=True,
        )


class UnifiedAnnotatorMetric(AnnotatorMetric):
    """Main class to compute metrics for a unified dataset.

    Example:
        >>> import argilla as rg
        >>> from argilla.client.feedback.metrics import UnifiedAnnotatorMetric
        >>> metric = UnifiedAnnotatorMetric(dataset=dataset, question_name=question)
        >>> metrics_report = metric.compute("accuracy")
    """

    def __init__(
        self,
        dataset: "FeedbackDataset",
        question_name: str,
        strategy_name: str = "majority",
        filter_by: Optional[Dict[str, Union["ResponseStatusFilter", List["ResponseStatusFilter"]]]] = None,
        sort_by: Optional[List["SortBy"]] = None,
        max_records: Optional[int] = None,
        responses_vs_suggestions: bool = True,
    ) -> None:
        self._metrics_per_question = METRICS_PER_QUESTION_UNIFIED
        super().__init__(dataset, question_name, responses_vs_suggestions=responses_vs_suggestions)
        self._filter_by = filter_by
        self._sort_by = sort_by
        self._max_records = max_records
        self._strategy_name = strategy_name

    def _check_responses_and_suggestions(
        self, unified_responses: "Responses", suggestions: "Suggestions"
    ) -> Tuple["Responses", "Suggestions"]:
        # Check for possible missing suggestions
        df_suggestions = pd.Series(suggestions)
        df_responses = pd.Series(unified_responses)
        df_responses = df_responses[df_suggestions.notna()]
        df_suggestions = df_suggestions[df_suggestions.notna()]
        total_responses = len(suggestions)

        unified_responses = df_responses.to_list()
        suggestions = df_suggestions.to_list()

        if len(suggestions) == 0:
            raise ValueError("All the suggestions are None, the metric cannot be computed.")
        elif len(suggestions) < total_responses:
            warnings.warn("Some suggestions are None, the metric will be computed without them.")
        return unified_responses, suggestions

    def compute(self, metric_names: Union[str, List[str]]) -> Union[ModelMetricResult, List[ModelMetricResult]]:
        """Computes the unified annotation metrics for the given question.

        Args:
            metric_names: name or list of names for the metrics to compute. i.e. `accuracy`
            kwargs: additional arguments to pass to the metric.

        Raises:
            ValueError: If the metric name is not supported for the given question.

        Returns:
            metrics: List of annotator metrics results if more than one metric is computed, or the result
                container if only one metric is computed.
        """
        metric_names = self._check_metrics(metric_names)
        metric_classes = self._get_metric_classes(metric_names)

        unified_responses, suggestions = get_unified_responses_and_suggestions(
            self._dataset,
            self._question_name,
            strategy_name=self._strategy_name,
            filter_by=self._filter_by,
            sort_by=self._sort_by,
            max_records=self._max_records,
        )
        self._check_responses_and_suggestions(unified_responses, suggestions)

        as_unified_responses, as_suggestions = self._prepare_responses_and_suggestions(unified_responses, suggestions)
        metrics = []
        for metric_name, metric_cls in metric_classes:
            metric = metric_cls(responses=as_unified_responses, suggestions=as_suggestions)
            result = metric.compute()
            metrics.append(ModelMetricResult(metric_name=metric_name, result=result, count=len(unified_responses)))

        if len(metric_names) == 1:
            return metrics[0]

        return metrics


class UnifiedModelMetric(UnifiedAnnotatorMetric):
    """"""

    def __init__(
        self,
        dataset: "FeedbackDataset",
        question_name: str,
        filter_by: Optional[Dict[str, Union["ResponseStatusFilter", List["ResponseStatusFilter"]]]] = None,
        sort_by: Optional[List["SortBy"]] = None,
        max_records: Optional[int] = None,
    ) -> None:
        super().__init__(
            dataset,
            question_name,
            filter_by=filter_by,
            sort_by=sort_by,
            max_records=max_records,
            responses_vs_suggestions=True,
        )


class AccuracyMetric(AnnotatorMetricBase):
    """Accuracy score.

    Which proportion of the responses are equal to the suggestions offered.

    We use the implementation in:
    https://scikit-learn.org/stable/modules/generated/sklearn.metrics.accuracy_score.html#sklearn.metrics.accuracy_score

    In multilabel classification, this function computes subset accuracy: the set of labels predicted for a
    sample must exactly match the corresponding set of labels in y_true
    """

    @requires_dependencies("scikit-learn")
    def _compute(self, responses, suggestions):
        from sklearn.metrics import accuracy_score

        return accuracy_score(y_true=responses, y_pred=suggestions)


class PrecisionMetric(AnnotatorMetricBase):
    """Compute the precision: tp / (tp + fp)

    We use the implementation in:
    https://scikit-learn.org/stable/modules/generated/sklearn.metrics.precision_score.html#sklearn.metrics.precision_score

    In case of multiclass data, calculate metrics for each label, and find their unweighted mean.
    This does not take label imbalance into account.
    """

    @requires_dependencies("scikit-learn")
    def _compute(self, responses, suggestions):
        from sklearn.metrics import precision_score

        if is_multiclass(responses) or is_multiclass(suggestions):
            kwargs = {"average": "macro"}
        else:
            kwargs = {"average": "binary", "pos_label": random.choice(np.unique(responses))}
        return precision_score(y_true=responses, y_pred=suggestions, **kwargs)


class RecallMetric(AnnotatorMetricBase):
    """Compute the recall: tp / (tp + fn)

    We use the implementation in:
    https://scikit-learn.org/stable/modules/generated/sklearn.metrics.recall_score.html#sklearn.metrics.recall_score

    In case of multiclass data, calculate metrics for each label, and find their unweighted mean.
    This does not take label imbalance into account.
    """

    @requires_dependencies("scikit-learn")
    def _compute(self, responses, suggestions):
        from sklearn.metrics import recall_score

        if is_multiclass(responses) or is_multiclass(suggestions):
            kwargs = {"average": "macro"}
        else:
            kwargs = {"average": "binary", "pos_label": random.choice(np.unique(responses))}
        return recall_score(y_true=responses, y_pred=suggestions, **kwargs)


class F1ScoreMetric(AnnotatorMetricBase):
    """F1 score: 2 * (precision * recall) / (precision + recall)

    We use the implementation in:
    https://scikit-learn.org/stable/modules/generated/sklearn.metrics.f1_score.html#sklearn.metrics.f1_score

    In case of multiclass data, calculate metrics for each label, and find their unweighted mean.
    This does not take label imbalance into account.
    """

    @requires_dependencies("scikit-learn")
    def _compute(self, responses, suggestions):
        from sklearn.metrics import f1_score

        if is_multiclass(responses) or is_multiclass(suggestions):
            kwargs = {"average": "macro"}
        else:
            kwargs = {"average": "binary", "pos_label": random.choice(np.unique(responses))}

        return f1_score(responses, suggestions, **kwargs)


class MultiLabelMetrics(AnnotatorMetricBase):
    """Parent class for MultiLabel based metrics. It binarizes the data to compute the metrics."""

    @requires_dependencies("scikit-learn")
    def _pre_process(self, responses, suggestions) -> Any:
        from sklearn.preprocessing import MultiLabelBinarizer

        classes = sorted(set(responses).union(set(suggestions)))
        # Keep the binarizer to access the classes later
        self._mlb = MultiLabelBinarizer()
        self._mlb.fit(classes)
        responses = self._mlb.transform(responses)
        suggestions = self._mlb.transform(suggestions)
        return responses, suggestions

    def _compute(self, responses, suggestions):
        # Child classes are in charge of the implementation
        pass


class MultiLabelAccuracyMetric(MultiLabelMetrics):
    """Computes the accuracy on the binarized data for multilabel classification.

    See Also:
        https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MultiLabelBinarizer.html
        `AccuracyMetric`
    """

    @requires_dependencies("scikit-learn")
    def _compute(self, responses, suggestions):
        from sklearn.metrics import accuracy_score

        return accuracy_score(y_true=responses, y_pred=suggestions)


class MultiLabelPrecisionMetric(MultiLabelMetrics):
    """Computes the precision on the binarized data for multilabel classification.

    See Also:
        https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MultiLabelBinarizer.html
        `PrecisionMetric`
    """

    @requires_dependencies("scikit-learn")
    def _compute(self, responses, suggestions):
        from sklearn.metrics import precision_score

        return precision_score(y_true=responses, y_pred=suggestions, average="macro")


class MultiLabelRecallMetric(MultiLabelMetrics):
    """Computes the recall on the binarized data for multilabel classification.

    See Also:
        https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MultiLabelBinarizer.html
        `RecallMetric`
    """

    @requires_dependencies("scikit-learn")
    def _compute(self, responses, suggestions):
        from sklearn.metrics import recall_score

        return recall_score(y_true=responses, y_pred=suggestions, average="macro")


class MultiLabelF1ScoreMetric(MultiLabelMetrics):
    """Computes the f1-score on the binarized data for multilabel classification.

    See Also:
        https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MultiLabelBinarizer.html
        `F1ScoreMetric`
    """

    @requires_dependencies("scikit-learn")
    def _compute(self, responses, suggestions):
        from sklearn.metrics import f1_score

        return f1_score(y_true=responses, y_pred=suggestions, average="macro")


class ConfusionMatrixMetric(AnnotatorMetricBase):
    """Compute confusion matrix to evaluate the accuracy of an annotator.

    In case of multiclass classification, this function returns a confusion matrix class-wise.
    """

    @requires_dependencies("scikit-learn")
    def _compute(self, responses, suggestions):
        import pandas as pd
        from sklearn.metrics import confusion_matrix

        unique_responses = sorted(np.unique(responses))
        unique_suggestions = sorted(np.unique(suggestions))
        labels = sorted(set(unique_responses).union(set(unique_suggestions)))
        labels_index = [f"responses_{label}" for label in labels]
        labels_columns = [f"suggestions_{label}" for label in labels]
        result = confusion_matrix(y_true=responses, y_pred=suggestions, labels=labels)
        return pd.DataFrame(result, index=labels_index, columns=labels_columns)


class MultiLabelConfusionMatrixMetric(MultiLabelMetrics):
    """Compute confusion matrix to evaluate the accuracy of an annotator.

    The data is binarized, so we will return a dict with the confusion matrix for each class.
    """

    @requires_dependencies("scikit-learn")
    def _compute(self, responses, suggestions):
        import pandas as pd
        from sklearn.metrics import multilabel_confusion_matrix

        unique_responses = sorted(np.unique(responses))
        unique_suggestions = sorted(np.unique(suggestions))
        labels = sorted(set(unique_responses).union(set(unique_suggestions)))
        matrices = multilabel_confusion_matrix(y_true=responses, y_pred=suggestions, labels=labels)
        report = {}
        for class_, matrix in zip(self._mlb.classes_, matrices):
            labels_index = [f"responses_{class_}_{i}" for i in ["true", "false"]]
            labels_columns = [f"suggestions_{class_}_{i}" for i in ["true", "false"]]
            report[class_] = pd.DataFrame(matrix, index=labels_index, columns=labels_columns)
        return report


class PearsonCorrelationCoefficientMetric(AnnotatorMetricBase):
    def _pre_process(self, responses, suggestions) -> Tuple[List[int], List[int]]:
        return map_str_to_int(responses), map_str_to_int(suggestions)

    @requires_dependencies("scipy")
    def _compute(self, responses, suggestions):
        import scipy.stats as stats

        return stats.pearsonr(x=suggestions, y=responses)[0]


class SpearmanCorrelationCoefficientMetric(AnnotatorMetricBase):
    @requires_dependencies("scipy")
    def _compute(self, responses, suggestions):
        import scipy.stats as stats

        return stats.spearmanr(a=suggestions, b=responses)[0]


class GLEUMetric(AnnotatorMetricBase):
    """
    Improvement of BLEU that takes into account the length of the response.

    BLEU (Bilingual Evaluation Understudy) is an algorithm for evaluating the quality of text
    which has been machine-translated from one natural language to another.
    The Google-BLEU is an improvement of BLEU that adresses some undesirable properties found on
    single sentences.

    https://huggingface.co/spaces/evaluate-metric/bleu
    https://huggingface.co/spaces/evaluate-metric/google_bleu
    """

    def _pre_process(self, responses, suggestions) -> Any:
        return responses, [[suggestion] for suggestion in suggestions]

    @requires_dependencies("evaluate")
    def _compute(self, responses: List[str], suggestions: List[str]):
        import evaluate

        gleu = evaluate.load("google_bleu")
        return gleu.compute(predictions=responses, references=suggestions)["google_bleu"]  # test is symmetrical


class ROUGEMetric(AnnotatorMetricBase):
    """
    From the evaluate library:

        ROUGE, or Recall-Oriented Understudy for Gisting Evaluation, is a set of metrics and a software package
        used for evaluating automatic summarization and machine translation software in natural language processing.
        The metrics compare an automatically produced summary or translation against a reference or a set of references
        (human-produced) summary or translation.
        Note that ROUGE is case insensitive, meaning that upper case letters are treated the same way as lower case letters.

    https://huggingface.co/spaces/evaluate-metric/rouge
    """

    @requires_dependencies("evaluate")
    def _compute(self, responses: List[str], suggestions: List[str]):
        import evaluate

        rouge = evaluate.load("rouge")
        return rouge.compute(predictions=responses, references=suggestions)  # test is symmetrical


class NDCGMetric(AnnotatorMetricBase):
    """Compute Normalized Discounted Cumulative Gain.

    From the Wikipedia page for Discounted Cumulative Gain:

    “Discounted cumulative gain (DCG) is a measure of ranking quality. In information retrieval,
    it is often used to measure effectiveness of web search engine algorithms or related applications.
    Using a graded relevance scale of documents in a search-engine result set, DCG measures the usefulness,
    or gain, of a document based on its position in the result list. The gain is accumulated from the
    top of the result list to the bottom, with the gain of each result discounted at lower ranks”

    See Also:
        https://scikit-learn.org/stable/modules/generated/sklearn.metrics.ndcg_score.html
        https://en.wikipedia.org/wiki/Discounted_cumulative_gain
    """

    @requires_dependencies("scikit-learn")
    def _compute(self, responses: List[str], suggestions: List[str]):
        from sklearn.metrics import ndcg_score

        return ndcg_score(y_true=responses, y_score=suggestions)


METRICS_PER_QUESTION = {
    LabelQuestion: {
        "accuracy": AccuracyMetric,
        "f1-score": F1ScoreMetric,
        "precision": PrecisionMetric,
        "recall": RecallMetric,
        "confusion-matrix": ConfusionMatrixMetric,
        "pearson-r": PearsonCorrelationCoefficientMetric,
    },
    MultiLabelQuestion: {
        "accuracy": MultiLabelAccuracyMetric,
        "f1-score": MultiLabelF1ScoreMetric,
        "precision": MultiLabelPrecisionMetric,
        "recall": MultiLabelRecallMetric,
        "confusion-matrix": MultiLabelConfusionMatrixMetric,
    },
    RatingQuestion: {
        "accuracy": AccuracyMetric,
        "f1-score": F1ScoreMetric,
        "precision": PrecisionMetric,
        "recall": RecallMetric,
        "confusion-matrix": ConfusionMatrixMetric,
        "spearman-r": SpearmanCorrelationCoefficientMetric,
    },
    TextQuestion: {
        "gleu": GLEUMetric,
        "rouge": ROUGEMetric,
    },
    RankingQuestion: {
        "ndcg-score": NDCGMetric,
    },
}


METRICS_PER_QUESTION_UNIFIED = {
    LabelQuestion: METRICS_PER_QUESTION[LabelQuestion],
    MultiLabelQuestion: METRICS_PER_QUESTION[MultiLabelQuestion],
    RatingQuestion: METRICS_PER_QUESTION[RatingQuestion],
    RankingQuestion: METRICS_PER_QUESTION[RankingQuestion],
}



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING

from argilla.client.feedback.training.base import ArgillaTrainerSkeleton
from argilla.training.openai import ArgillaOpenAITrainer as ArgillaOpenAITrainerV1
from argilla.utils.dependency import require_dependencies

if TYPE_CHECKING:
    from argilla.client.feedback.integrations.huggingface.model_card import OpenAIModelCardData


class ArgillaOpenAITrainer(ArgillaOpenAITrainerV1, ArgillaTrainerSkeleton):
    def __init__(self, *args, **kwargs) -> None:
        require_dependencies("openai>=0.27.10")
        ArgillaTrainerSkeleton.__init__(self, *args, **kwargs)

        self.__legacy = False
        self.sleep_timer = 10
        self.device = None
        self.finetune_id = None

        if self._seed is not None:
            self._logger.warning("Seed is not supported for OpenAI. Ignoring seed for training.")

        if self._model is None:
            self._model = "gpt-3.5-turbo-0613"

        if isinstance(self._dataset, tuple):
            self._train_dataset = self._dataset[0]
            self._eval_dataset = self._dataset[1]
        else:
            self._train_dataset = self._dataset
            self._eval_dataset = None

        if self._model != "gpt-3.5-turbo-0613":
            raise NotImplementedError("Legacy models are not supported for OpenAI with the FeedbackDataset.")

        self.init_training_args(model=self._model)

    def get_model_card_data(self, **card_data_kwargs) -> "OpenAIModelCardData":
        """
        Generate the card data to be used for the `ArgillaModelCard`.

        Args:
            card_data_kwargs: Extra arguments provided by the user when creating the `ArgillaTrainer`.

        Returns:
            OpenAIModelCardData: Container for the data to be written on the `ArgillaModelCard`.
        """
        from argilla.client.feedback.integrations.huggingface.model_card import OpenAIModelCardData

        return OpenAIModelCardData(
            model_name=self._model,
            task=self._task,
            **card_data_kwargs,
        )

    def push_to_huggingface(self, repo_id: str, **kwargs) -> None:
        raise NotImplementedError("This method is not implemented for `ArgillaOpenAITrainer`.")



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from functools import lru_cache
from typing import List, Optional, Union

import httpx

from argilla._constants import WORKSPACE_HEADER_NAME
from argilla.client.sdk.client import AuthenticatedClient
from argilla.client.sdk.commons.errors_handler import handle_response_error
from argilla.client.sdk.commons.models import Response
from argilla.client.sdk.datasets.models import CopyDatasetRequest, Dataset


@lru_cache(maxsize=None)
def get_dataset(client: AuthenticatedClient, name: str, workspace: Optional[str] = None) -> Response[Dataset]:
    url = f"{client.base_url}/api/datasets/{name}"

    params = {"workspace": workspace} if workspace else None

    response = httpx.get(
        url=url,
        params=params,
        headers=client.get_headers(),
        cookies=client.get_cookies(),
        timeout=client.get_timeout(),
    )

    if response.status_code == 200:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = Dataset(**response.json())
        return response_obj

    handle_response_error(response)


def list_datasets(client: AuthenticatedClient, workspace: Optional[str] = None) -> Response[List[Dataset]]:
    url = f"{client.base_url}/api/datasets"

    headers = client.get_headers().copy()
    headers.pop(WORKSPACE_HEADER_NAME, None)

    response = httpx.get(
        url=url,
        params={"workspace": workspace} if workspace else None,
        headers=headers,
        cookies=client.get_cookies(),
        timeout=client.get_timeout(),
    )

    if response.status_code == 200:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = [Dataset(**dataset) for dataset in response.json()]
        return response_obj
    handle_response_error(response)


def copy_dataset(client: AuthenticatedClient, name: str, json_body: CopyDatasetRequest) -> Response[Dataset]:
    url = f"{client.base_url}/api/datasets/{name}:copy"

    response = httpx.put(
        url=url,
        headers=client.get_headers(),
        cookies=client.get_cookies(),
        timeout=client.get_timeout(),
        json=json_body.dict(by_alias=True),
    )

    if response.status_code == 200:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = Dataset(**response.json())
        return response_obj
    handle_response_error(response)


def delete_dataset(client: AuthenticatedClient, name: str) -> Response:
    url = f"{client.base_url}/api/datasets/{name}"

    response = httpx.delete(
        url=url,
        headers=client.get_headers(),
        cookies=client.get_cookies(),
        timeout=client.get_timeout(),
    )
    # If everything was ok, we must clear local cache data
    if 200 <= response.status_code < 400:
        get_dataset.cache_clear()
        return Response(
            status_code=response.status_code,
            content=response.content,
            headers=response.headers,
            parsed=response.json(),
        )
    handle_response_error(response, dataset=name)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import TYPE_CHECKING, List, Literal, Optional, Type, Union
from uuid import UUID

from argilla.client.feedback.dataset.helpers import get_dataset_by_name_and_workspace
from argilla.client.feedback.dataset.remote.dataset import INCLUDE_ALL_VECTORS_PARAM, RemoteFeedbackDataset
from argilla.client.feedback.schemas.enums import FieldTypes, MetadataPropertyTypes, QuestionTypes
from argilla.client.feedback.schemas.fields import TextField
from argilla.client.feedback.schemas.questions import (
    LabelQuestion,
    MultiLabelQuestion,
    RankingQuestion,
    RatingQuestion,
    TextQuestion,
)
from argilla.client.feedback.schemas.remote.fields import RemoteTextField
from argilla.client.feedback.schemas.remote.metadata import (
    RemoteFloatMetadataProperty,
    RemoteIntegerMetadataProperty,
    RemoteTermsMetadataProperty,
)
from argilla.client.feedback.schemas.remote.questions import (
    RemoteLabelQuestion,
    RemoteMultiLabelQuestion,
    RemoteRankingQuestion,
    RemoteRatingQuestion,
    RemoteTextQuestion,
)
from argilla.client.feedback.schemas.types import AllowedMetadataPropertyTypes
from argilla.client.feedback.schemas.vector_settings import VectorSettings
from argilla.client.sdk.commons.errors import AlreadyExistsApiError
from argilla.client.sdk.v1.datasets import api as datasets_api_v1
from argilla.client.singleton import ArgillaSingleton
from argilla.client.workspaces import Workspace

if TYPE_CHECKING:
    import httpx

    from argilla.client.client import Argilla as ArgillaClient
    from argilla.client.feedback.dataset.local import FeedbackDataset
    from argilla.client.feedback.dataset.local.dataset import FeedbackDataset
    from argilla.client.feedback.schemas.types import (
        AllowedFieldTypes,
        AllowedMetadataPropertyTypes,
        AllowedQuestionTypes,
        AllowedRemoteFieldTypes,
        AllowedRemoteMetadataPropertyTypes,
        AllowedRemoteQuestionTypes,
    )
    from argilla.client.sdk.v1.datasets.models import (
        FeedbackDatasetModel,
        FeedbackFieldModel,
        FeedbackMetadataPropertyModel,
        FeedbackQuestionModel,
    )

_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.INFO)


def _prepare_workspace(client, workspace):
    if workspace is None:
        workspace = client.get_workspace()

    if workspace is None:
        raise ValueError("No workspace provided and no active workspace found.")

    if isinstance(workspace, str):
        workspace = Workspace.from_name(workspace)

    return workspace


class ArgillaMixin:
    @staticmethod
    def __delete_dataset(client: "httpx.Client", id: UUID) -> None:
        try:
            datasets_api_v1.delete_dataset(client=client, id=id)
        except Exception as e:
            raise Exception(
                f"Failed while deleting the `FeedbackDataset` with ID '{id}' from Argilla with" f" exception: {e}"
            ) from e

    @staticmethod
    def _parse_to_remote_field(field: "FeedbackFieldModel") -> "AllowedRemoteFieldTypes":
        if field.settings["type"] == FieldTypes.text:
            field = RemoteTextField.from_api(field)
        else:
            raise ValueError(
                f"Field '{field.name}' is not a supported field in the current Python package version,"
                f" supported field types are: `{'`, `'.join([arg.value for arg in FieldTypes])}`."
            )
        return field

    @staticmethod
    def __add_fields(
        fields: List["AllowedFieldTypes"], client: "httpx.Client", id: UUID
    ) -> List["AllowedRemoteFieldTypes"]:
        uploaded_fields = []
        for field in fields:
            try:
                new_field = datasets_api_v1.add_field(client=client, id=id, field=field.to_server_payload()).parsed
                uploaded_fields.append(ArgillaMixin._parse_to_remote_field(new_field))
            except Exception as e:
                ArgillaMixin.__delete_dataset(client=client, id=id)
                raise Exception(
                    f"Failed while adding the field '{field.name}' to the `FeedbackDataset` in Argilla with"
                    f" exception: {e}"
                ) from e
        return uploaded_fields

    @staticmethod
    def _parse_to_remote_question(question: "FeedbackQuestionModel") -> "AllowedRemoteQuestionTypes":
        if question.settings["type"] == QuestionTypes.rating:
            question = RemoteRatingQuestion.from_api(question)
        elif question.settings["type"] == QuestionTypes.text:
            question = RemoteTextQuestion.from_api(question)
        elif question.settings["type"] == QuestionTypes.label_selection:
            question = RemoteLabelQuestion.from_api(question)
        elif question.settings["type"] == QuestionTypes.multi_label_selection:
            question = RemoteMultiLabelQuestion.from_api(question)
        elif question.settings["type"] == QuestionTypes.ranking:
            question = RemoteRankingQuestion.from_api(question)
        else:
            raise ValueError(
                f"Question '{question.name}' is not a supported question in the current Python package"
                f" version, supported question types are: `{'`, `'.join([arg.value for arg in QuestionTypes])}`."
            )

        return question

    @staticmethod
    def __add_questions(
        questions: List["AllowedQuestionTypes"], client: "httpx.Client", id: UUID
    ) -> List["AllowedRemoteQuestionTypes"]:
        uploaded_questions = []
        for question in questions:
            try:
                new_question = datasets_api_v1.add_question(
                    client=client, id=id, question=question.to_server_payload()
                ).parsed
                uploaded_questions.append(ArgillaMixin._parse_to_remote_question(new_question))
            except Exception as e:
                ArgillaMixin.__delete_dataset(client=client, id=id)
                raise Exception(
                    f"Failed while adding the question '{question.name}' to the `FeedbackDataset` in Argilla with"
                    f" exception: {e}"
                ) from e
        return uploaded_questions

    @staticmethod
    def _parse_to_remote_metadata_property(
        metadata_property: "FeedbackMetadataPropertyModel",
        client: Optional["httpx.Client"] = None,
    ) -> "AllowedRemoteMetadataPropertyTypes":
        if metadata_property.settings["type"] == MetadataPropertyTypes.terms:
            metadata_property = RemoteTermsMetadataProperty.from_api(payload=metadata_property, client=client)
        elif metadata_property.settings["type"] == MetadataPropertyTypes.integer:
            metadata_property = RemoteIntegerMetadataProperty.from_api(payload=metadata_property, client=client)
        elif metadata_property.settings["type"] == MetadataPropertyTypes.float:
            metadata_property = RemoteFloatMetadataProperty.from_api(payload=metadata_property, client=client)
        else:
            raise ValueError(
                f"Metadata property '{metadata_property.name}' is not a supported metadata property in the current"
                " Python package version, supported metadata property types are: "
                f"`{'`, `'.join([arg.value for arg in MetadataPropertyTypes])}`."
            )

        return metadata_property

    @staticmethod
    def __add_metadata_properties(
        metadata_properties: List["AllowedMetadataPropertyTypes"], client: "httpx.Client", id: UUID
    ) -> Union[List["AllowedRemoteMetadataPropertyTypes"], None]:
        if not metadata_properties:
            return None

        uploaded_metadata_properties = []
        for metadata_property in metadata_properties:
            try:
                new_metadata_property = datasets_api_v1.add_metadata_property(
                    client=client, id=id, metadata_property=metadata_property.to_server_payload()
                ).parsed
                uploaded_metadata_properties.append(
                    ArgillaMixin._parse_to_remote_metadata_property(
                        metadata_property=new_metadata_property, client=client
                    )
                )
            except Exception as e:
                ArgillaMixin.__delete_dataset(client=client, id=id)
                raise Exception(
                    f"Failed while adding the metadata property '{metadata_property.name}' to the `FeedbackDataset` in"
                    f" Argilla with exception: {e}"
                ) from e
        return uploaded_metadata_properties

    @staticmethod
    def __publish_dataset(client: "httpx.Client", id: UUID) -> None:
        try:
            datasets_api_v1.publish_dataset(client=client, id=id)
        except Exception as e:
            ArgillaMixin.__delete_dataset(client=client, id=id)
            raise Exception(f"Failed while publishing the `FeedbackDataset` in Argilla with exception: {e}") from e

    def push_to_argilla(
        self: Union["FeedbackDataset", "ArgillaMixin"],
        name: str,
        workspace: Optional[Union[str, Workspace]] = None,
        show_progress: bool = True,
    ) -> RemoteFeedbackDataset:
        """Pushes the `FeedbackDataset` to Argilla.

        Note that you may need to `rg.init(...)` with your Argilla credentials before calling this function, otherwise
        the default http://localhost:6900 will be used, which will fail if Argilla is not deployed locally.

        Args:
            name: the name of the dataset to push to Argilla.
            workspace: the workspace where to push the dataset to. If not provided, the active workspace will be used.
            show_progress: the option to choose to show/hide tqdm progress bar while looping over records.

        Returns:
            The `FeedbackDataset` pushed to Argilla, which is now an instance of `RemoteFeedbackDataset`.
        """
        client: "ArgillaClient" = ArgillaSingleton.get()
        httpx_client: "httpx.Client" = client.http_client.httpx

        workspace = _prepare_workspace(client, workspace)
        created_dataset = _create_argilla_dataset_or_raise(httpx_client, name, workspace, dataset=self)

        try:
            ArgillaMixin.__add_fields(fields=self.fields, client=httpx_client, id=created_dataset.id)
            ArgillaMixin.__add_questions(questions=self.questions, client=httpx_client, id=created_dataset.id)

            if self.metadata_properties:
                ArgillaMixin.__add_metadata_properties(
                    metadata_properties=self.metadata_properties, client=httpx_client, id=created_dataset.id
                )

            if self.vectors_settings:
                ArgillaMixin.__add_vectors_settings(
                    vectors_settings=self.vectors_settings, client=httpx_client, id=created_dataset.id
                )

            ArgillaMixin.__publish_dataset(client=httpx_client, id=created_dataset.id)

            # TODO: Remote dataset should connect all settings by API calls requested on demand.
            #  Once is done, this prefetch info should be removed.
            fields = ArgillaMixin.__get_fields(client=httpx_client, id=created_dataset.id)
            questions = ArgillaMixin.__get_questions(client=httpx_client, id=created_dataset.id)

            remote_dataset = RemoteFeedbackDataset(
                client=httpx_client,
                id=created_dataset.id,
                name=name,
                workspace=workspace,
                created_at=created_dataset.inserted_at,
                updated_at=created_dataset.updated_at,
                fields=fields,
                questions=questions,
                guidelines=self.guidelines,
                allow_extra_metadata=self.allow_extra_metadata,
            )

            if len(self.records) > 0:
                remote_dataset.add_records(self.records, show_progress)

            _LOGGER.info("✓ Dataset succesfully pushed to Argilla")
            _LOGGER.info(remote_dataset)

            return remote_dataset
        except Exception as ex:
            ArgillaMixin.__delete_dataset(client=httpx_client, id=created_dataset.id)
            raise ex

    @staticmethod
    def __get_fields(client: "httpx.Client", id: UUID) -> List["AllowedRemoteFieldTypes"]:
        fields = []
        for field in datasets_api_v1.get_fields(client=client, id=id).parsed:
            fields.append(ArgillaMixin._parse_to_remote_field(field))
        return fields

    @staticmethod
    def __get_questions(client: "httpx.Client", id: UUID) -> List["AllowedRemoteQuestionTypes"]:
        questions = []
        for question in datasets_api_v1.get_questions(client=client, id=id).parsed:
            questions.append(ArgillaMixin._parse_to_remote_question(question))
        return questions

    @staticmethod
    def __get_metadata_properties(client: "httpx.Client", id: UUID) -> List["AllowedRemoteMetadataPropertyTypes"]:
        metadata_properties = []
        for metadata_prop in datasets_api_v1.get_metadata_properties(client=client, id=id).parsed:
            metadata_properties.append(
                ArgillaMixin._parse_to_remote_metadata_property(metadata_property=metadata_prop, client=client)
            )
        return metadata_properties

    @classmethod
    def from_argilla(
        cls: Type["FeedbackDataset"],
        name: Optional[str] = None,
        *,
        workspace: Optional[str] = None,
        id: Optional[Union[UUID, str]] = None,
        with_vectors: Union[Literal[INCLUDE_ALL_VECTORS_PARAM], List[str], None] = None,
    ) -> RemoteFeedbackDataset:
        """Retrieves an existing `FeedbackDataset` from Argilla (must have been pushed in advance).

        Note that even though no argument is mandatory, you must provide either the `name`,
        the combination of `name` and `workspace`, or the `id`, otherwise an error will be raised.

        Args:
            name: the name of the `FeedbackDataset` to retrieve from Argilla. Defaults to `None`.
            workspace: the workspace of the `FeedbackDataset` to retrieve from Argilla.
                If not provided, the active workspace will be used.
            id: the ID of the `FeedbackDataset` to retrieve from Argilla. Defaults to `None`.
            with_vectors: the vector settings to retrieve from Argilla. Use `all` to download all vectors.
                Defaults to `None`.

        Returns:
            The `RemoteFeedbackDataset` retrieved from Argilla.

        Raises:
            ValueError: if no `FeedbackDataset` with the provided `name`, `workspace`, or `id` exists in Argilla.

        Examples:
            >>> import argilla as rg
            >>> rg.init(...)
            >>> dataset = rg.FeedbackDataset.from_argilla(name="my_dataset")
        """
        httpx_client: "httpx.Client" = ArgillaSingleton.get().http_client.httpx

        existing_dataset = get_dataset_by_name_and_workspace(name=name, workspace=workspace, id=id)
        if existing_dataset is None:
            raise ValueError(
                f"Could not find a `FeedbackDataset` in Argilla with name='{name}'."
                if name and not workspace
                else (
                    f"Could not find a `FeedbackDataset` in Argilla with name='{name}' and workspace='{workspace}'."
                    if name and workspace
                    else f"Could not find a `FeedbackDataset` in Argilla with ID='{id}'."
                )
            )

        fields = ArgillaMixin.__get_fields(client=httpx_client, id=existing_dataset.id)
        questions = ArgillaMixin.__get_questions(client=httpx_client, id=existing_dataset.id)

        return RemoteFeedbackDataset(
            client=httpx_client,
            id=existing_dataset.id,
            name=existing_dataset.name,
            workspace=Workspace.from_id(existing_dataset.workspace_id),
            created_at=existing_dataset.inserted_at,
            updated_at=existing_dataset.updated_at,
            fields=fields,
            questions=questions,
            guidelines=existing_dataset.guidelines or None,
            allow_extra_metadata=existing_dataset.allow_extra_metadata,
            with_vectors=with_vectors,
        )

    @classmethod
    def list(cls: Type["FeedbackDataset"], workspace: Optional[str] = None) -> List[RemoteFeedbackDataset]:
        """Lists the `FeedbackDataset`s pushed to Argilla.

        Note that you may need to `rg.init(...)` with your Argilla credentials before
        calling this function, otherwise, the default http://localhost:6900 will be used,
        which will fail if Argilla is not deployed locally.

        Args:
            workspace: the workspace where to list the datasets from. If not provided,
                then the workspace filtering won't be applied. Defaults to `None`.

        Returns:
            A list of `RemoteFeedbackDataset` datasets, which are `FeedbackDataset`
            datasets previously pushed to Argilla via `push_to_argilla`.
        """
        client: "ArgillaClient" = ArgillaSingleton.get()
        httpx_client: "httpx.Client" = client.http_client.httpx

        if workspace is not None:
            workspace = Workspace.from_name(workspace)

        try:
            datasets = datasets_api_v1.list_datasets(
                client=httpx_client, workspace_id=workspace.id if workspace is not None else None
            ).parsed
        except Exception as e:
            raise RuntimeError(
                f"Failed while listing the `FeedbackDataset` datasets in Argilla with exception: {e}"
            ) from e

        if len(datasets) == 0:
            return []
        return [
            RemoteFeedbackDataset(
                client=httpx_client,
                id=dataset.id,
                name=dataset.name,
                workspace=workspace if workspace is not None else Workspace.from_id(dataset.workspace_id),
                created_at=dataset.inserted_at,
                updated_at=dataset.updated_at,
                fields=ArgillaMixin.__get_fields(client=httpx_client, id=dataset.id),
                questions=ArgillaMixin.__get_questions(client=httpx_client, id=dataset.id),
                guidelines=dataset.guidelines or None,
            )
            for dataset in datasets
        ]

    @staticmethod
    def __add_vectors_settings(
        vectors_settings: Union[List[VectorSettings], None], client: "httpx.Client", id: UUID
    ) -> None:
        try:
            for vector_settings in vectors_settings or []:
                try:
                    datasets_api_v1.add_vector_settings(
                        client=client,
                        id=id,
                        name=vector_settings.name,
                        title=vector_settings.name,
                        dimensions=vector_settings.dimensions,
                    ).parsed
                except AlreadyExistsApiError:
                    raise ValueError(f"Vector settings with name {vector_settings.name!r} already exists.")
        except Exception as e:
            ArgillaMixin.__delete_dataset(client=client, id=id)
            raise Exception(f"Failed adding vectors to the `FeedbackDataset` in Argilla with exception: {e}") from e


def _create_argilla_dataset_or_raise(
    httpx_client: "httpx.Client", name: str, workspace: "Workspace", dataset: "FeedbackDataset"
) -> "FeedbackDatasetModel":
    argilla_dataset = get_dataset_by_name_and_workspace(name=name, workspace=workspace)

    if argilla_dataset is not None:
        raise RuntimeError(
            f"Dataset with name=`{name}` and workspace=`{workspace.name}` already exists in Argilla, please"
            " choose another name and/or workspace."
        )

    try:
        return datasets_api_v1.create_dataset(
            client=httpx_client,
            name=name,
            workspace_id=workspace.id,
            guidelines=dataset.guidelines,
            allow_extra_metadata=dataset.allow_extra_metadata,
        ).parsed
    except Exception as e:
        raise Exception(f"Failed while creating the `FeedbackDataset` in Argilla with exception: {e}") from e


class TaskTemplateMixin:
    """
    Mixin to add task template functionality to a `FeedbackDataset`.
    The NLP tasks covered are:
        "text_classification"
        "extractive_question_answering"
        "summarization"
        "translation"
        "sentence_similarity"
        "natural_language_inference"
        "supervised_fine_tuning"
        "preference_modeling/reward_modeling"
        "proximal_policy_optimization"
        "direct_preference_optimization"
        "retrieval_augmented_generation"
        "multi_modal_classification"
        "multi_modal_transcription"
    """

    @classmethod
    def for_text_classification(
        cls: Type["FeedbackDataset"],
        labels: List[str],
        multi_label: bool = False,
        use_markdown: bool = False,
        guidelines: Optional[str] = None,
        metadata_properties: List[AllowedMetadataPropertyTypes] = None,
        vectors_settings: List[VectorSettings] = None,
    ) -> "FeedbackDataset":
        """
        You can use this method to create a basic dataset for text classification tasks.

        Args:
            labels: A list of labels for your dataset
            multi_label: Set this parameter to True if you want to add multiple labels to your dataset
            use_markdown: Set this parameter to True if you want to use markdown in your dataset
            guidelines: Contains the guidelines for the dataset
            metadata_properties: Contains the metadata properties that will be indexed and could be used to filter the dataset. Defaults to `None`.
            vectors_settings: Define the configuration of the vectors associated to the records that will be used to perform the vector search. Defaults to `None`.

        Returns:
            A `FeedbackDataset` object for text classification containing "text" field and LabelQuestion or MultiLabelQuestion named "label"
        """
        default_guidelines = "This is a text classification dataset that contains texts and labels. Given a set of texts and a predefined set of labels, the goal of text classification is to assign one or more labels to each text based on its content. Please classify the texts by making the correct selection."

        description = "Classify the text by selecting the correct label from the given list of labels."
        return cls(
            fields=[TextField(name="text", use_markdown=use_markdown)],
            questions=[
                LabelQuestion(
                    name="label",
                    labels=labels,
                    description=description,
                )
                if not multi_label
                else MultiLabelQuestion(
                    name="label",
                    labels=labels,
                    description=description,
                )
            ],
            guidelines=guidelines
            if guidelines is not None
            else default_guidelines
            if multi_label
            else default_guidelines.replace("one or more labels", "one label"),
            metadata_properties=metadata_properties,
            vectors_settings=vectors_settings,
        )

    @classmethod
    def for_question_answering(
        cls: Type["FeedbackDataset"],
        use_markdown: bool = False,
        guidelines: Optional[str] = None,
        metadata_properties: List[AllowedMetadataPropertyTypes] = None,
        vectors_settings: List[VectorSettings] = None,
    ) -> "FeedbackDataset":
        """
        You can use this method to create a basic dataset for question answering tasks.

        Args:
            use_markdown: Set this parameter to True if you want to use markdown in your dataset
            guidelines: Contains the guidelines for the dataset
            metadata_properties: Contains the metadata properties that will be indexed and could be used to filter the dataset. Defaults to `None`.
            vectors_settings: Define the configuration of the vectors associated to the records that will be used to perform the vector search. Defaults to `None`.

        Returns:
            A `FeedbackDataset` object for question answering containing "context" and "question" fields and a TextQuestion named "answer"
        """
        default_guidelines = "This is a question answering dataset that contains questions and contexts. Please answer the question by using the context."
        return cls(
            fields=[
                TextField(name="question", use_markdown=use_markdown),
                TextField(name="context", use_markdown=use_markdown),
            ],
            questions=[
                TextQuestion(
                    name="answer",
                    description="Answer the question. Note that the answer must exactly be in the context.",
                    use_markdown=use_markdown,
                    required=True,
                )
            ],
            guidelines=default_guidelines if guidelines is None else guidelines,
            metadata_properties=metadata_properties,
            vectors_settings=vectors_settings,
        )

    @classmethod
    def for_summarization(
        cls: Type["FeedbackDataset"],
        use_markdown: bool = False,
        guidelines: Optional[str] = None,
        metadata_properties: List[AllowedMetadataPropertyTypes] = None,
        vectors_settings: List[VectorSettings] = None,
    ) -> "FeedbackDataset":
        """
        You can use this method to create a basic dataset for summarization tasks.

        Args:
            use_markdown: Set this parameter to True if you want to use markdown in your dataset
            guidelines: Contains the guidelines for the dataset
            metadata_properties: Contains the metadata properties that will be indexed and could be used to filter the dataset. Defaults to `None`.
            vectors_settings: Define the configuration of the vectors associated to the records that will be used to perform the vector search. Defaults to `None`.

        Returns:
            A `FeedbackDataset` object for summarization containing "text" field and a TextQuestion named "summary"
        """
        default_guidelines = (
            "This is a summarization dataset that contains texts. Please summarize the text in the text field."
        )
        return cls(
            fields=[TextField(name="text", use_markdown=use_markdown)],
            questions=[
                TextQuestion(name="summary", description="Write a summary of the text.", use_markdown=use_markdown)
            ],
            guidelines=default_guidelines if guidelines is None else guidelines,
            metadata_properties=metadata_properties,
            vectors_settings=vectors_settings,
        )

    @classmethod
    def for_translation(
        cls: Type["FeedbackDataset"],
        use_markdown: bool = False,
        guidelines: Optional[str] = None,
        metadata_properties: List[AllowedMetadataPropertyTypes] = None,
        vectors_settings: List[VectorSettings] = None,
    ) -> "FeedbackDataset":
        """
        You can use this method to create a basic dataset for translation tasks.

        Args:
            use_markdown: Set this parameter to True if you want to use markdown in your dataset
            guidelines: Contains the guidelines for the dataset
            metadata_properties: Contains the metadata properties that will be indexed and could be used to filter the dataset. Defaults to `None`.
            vectors_settings: Define the configuration of the vectors associated to the records that will be used to perform the vector search. Defaults to `None`.

        Returns:
            A `FeedbackDataset` object for translation containing "source" field and a TextQuestion named "target"
        """
        default_guidelines = (
            "This is a translation dataset that contains texts. Please translate the text in the text field."
        )
        return cls(
            fields=[TextField(name="source", use_markdown=use_markdown)],
            questions=[TextQuestion(name="target", description="Translate the text.", use_markdown=use_markdown)],
            guidelines=default_guidelines if guidelines is None else guidelines,
            metadata_properties=metadata_properties,
            vectors_settings=vectors_settings,
        )

    @classmethod
    def for_sentence_similarity(
        cls: Type["FeedbackDataset"],
        rating_scale: int = 7,
        use_markdown: bool = False,
        guidelines: Optional[str] = None,
        metadata_properties: List[AllowedMetadataPropertyTypes] = None,
        vectors_settings: List[VectorSettings] = None,
    ) -> "FeedbackDataset":
        """
        You can use this method to create a basic dataset for sentence similarity tasks.

        Args:
            rating_scale: Set this parameter to the number of similarity scale you want to add to your dataset
            use_markdown: Set this parameter to True if you want to use markdown in your dataset
            guidelines: Contains the guidelines for the dataset
            metadata_properties: Contains the metadata properties that will be indexed and could be used to filter the dataset. Defaults to `None`.
            vectors_settings: Define the configuration of the vectors associated to the records that will be used to perform the vector search. Defaults to `None`.

        Returns:
            A `FeedbackDataset` object for sentence similarity containing "sentence1" and "sentence2" fields and a RatingQuestion named "similarity"
        """
        default_guidelines = "This is a sentence similarity dataset that contains two sentences. Please rate the similarity between the two sentences."
        return cls(
            fields=[
                TextField(name="sentence-1", use_markdown=use_markdown),
                TextField(name="sentence-2", use_markdown=use_markdown),
            ],
            questions=[
                RatingQuestion(
                    name="similarity",
                    values=list(range(1, rating_scale + 1)),
                    description="Rate the similarity between the two sentences.",
                )
            ],
            guidelines=default_guidelines if guidelines is None else guidelines,
            metadata_properties=metadata_properties,
            vectors_settings=vectors_settings,
        )

    @classmethod
    def for_natural_language_inference(
        cls: Type["FeedbackDataset"],
        labels: Optional[List[str]] = None,
        use_markdown: bool = False,
        guidelines: Optional[str] = None,
        metadata_properties: List[AllowedMetadataPropertyTypes] = None,
        vectors_settings: List[VectorSettings] = None,
    ) -> "FeedbackDataset":
        """
        You can use this method to create a basic dataset for natural language inference tasks.

        Args:
            labels: A list of labels for your dataset
            use_markdown: Set this parameter to True if you want to use markdown in your dataset
            guidelines: Contains the guidelines for the dataset
            metadata_properties: Contains the metadata properties that will be indexed and could be used to filter the dataset. Defaults to `None`.
            vectors_settings: Define the configuration of the vectors associated to the records that will be used to perform the vector search. Defaults to `None`.

        Returns:
            A `FeedbackDataset` object for natural language inference containing "premise" and "hypothesis" fields and a LabelQuestion named "label"
        """
        default_guidelines = "This is a natural language inference dataset that contains premises and hypotheses. Please choose the correct label for the given premise and hypothesis."
        if labels is None:
            labels = ["entailment", "neutral", "contradiction"]
        return cls(
            fields=[
                TextField(name="premise", use_markdown=use_markdown),
                TextField(name="hypothesis", use_markdown=use_markdown),
            ],
            questions=[LabelQuestion(name="label", labels=labels, description="Choose one of the labels.")],
            guidelines=default_guidelines if guidelines is None else guidelines,
            metadata_properties=metadata_properties,
            vectors_settings=vectors_settings,
        )

    @classmethod
    def for_supervised_fine_tuning(
        cls: Type["FeedbackDataset"],
        context: bool = False,
        use_markdown: bool = False,
        guidelines: Optional[str] = None,
        metadata_properties: List[AllowedMetadataPropertyTypes] = None,
        vectors_settings: List[VectorSettings] = None,
    ) -> "FeedbackDataset":
        """
        You can use this method to create a basic dataset for supervised fine-tuning tasks.

        Args:
            context: Set this parameter to True if you want to add context to your dataset
            use_markdown: Set this parameter to True if you want to use markdown in your dataset
            guidelines: Contains the guidelines for the dataset
            metadata_properties: Contains the metadata properties that will be indexed and could be used to filter the dataset. Defaults to `None`.
            vectors_settings: Define the configuration of the vectors associated to the records that will be used to perform the vector search. Defaults to `None`.

        Returns:
            A `FeedbackDataset` object for supervised fine-tuning containing "instruction" and optional "context" field and a TextQuestion named "response"
        """
        default_guidelines = "This is a supervised fine-tuning dataset that contains instructions. Please write the response to the instruction in the response field."
        fields = [
            TextField(name="prompt", use_markdown=use_markdown),
        ]
        if context:
            fields.append(TextField(name="context", use_markdown=use_markdown, required=False))
        return cls(
            fields=fields,
            questions=[
                TextQuestion(
                    name="response", description="Write the response to the instruction.", use_markdown=use_markdown
                )
            ],
            guidelines=guidelines
            if guidelines is not None
            else default_guidelines + " Take the context into account when writing the response."
            if context
            else default_guidelines,
            metadata_properties=metadata_properties,
            vectors_settings=vectors_settings,
        )

    @classmethod
    def for_preference_modeling(
        cls: Type["FeedbackDataset"],
        number_of_responses: int = 2,
        context: bool = False,
        use_markdown: bool = False,
        guidelines: Optional[str] = None,
        metadata_properties: List[AllowedMetadataPropertyTypes] = None,
        vectors_settings: List[VectorSettings] = None,
    ) -> "FeedbackDataset":
        """
        You can use this method to create a basic dataset for preference tasks.

        Args:
            number_of_responses: Set this parameter to the number of responses you want to add to your dataset
            context: Set this parameter to True if you want to add context to your dataset
            use_markdown: Set this parameter to True if you want to use markdown in your dataset
            guidelines: Contains the guidelines for the dataset.
            metadata_properties: Contains the metadata properties that will be indexed and could be used to filter the dataset. Defaults to `None`.
            vectors_settings: Define the configuration of the vectors associated to the records that will be used to perform the vector search. Defaults to `None`.

        Returns:
            A `FeedbackDataset` object for preference containing "prompt", "option1" and "option2" fields and a RatingQuestion named "preference"
        """
        default_guidelines = "This is a preference dataset that contains contexts and options. Please rank the options that you would prefer in the given context."
        response_fields = [
            TextField(
                name="response" + str(response + 1),
                title="Response " + str(response + 1),
                use_markdown=use_markdown,
                required=True if response == 0 else False,
            )
            for response in range(number_of_responses)
        ]

        fields = [
            TextField(name="prompt", use_markdown=use_markdown),
        ] + response_fields

        if context:
            fields.insert(1, TextField(name="context", use_markdown=use_markdown, required=False))

        return cls(
            fields=fields,
            questions=[
                RankingQuestion(
                    name="preference",
                    title="Order responses based on your preference",
                    description=f"1 = Best, {number_of_responses} = Worst. Ties are allowed.",
                    required=True,
                    values={
                        "response" + str(response + 1): "Response " + str(response + 1)
                        for response in range(number_of_responses)
                    },
                )
            ],
            guidelines=default_guidelines if guidelines is None else guidelines,
            metadata_properties=metadata_properties,
            vectors_settings=vectors_settings,
        )

    @classmethod
    def for_proximal_policy_optimization(
        cls: Type["FeedbackDataset"],
        rating_scale: int = 7,
        context: bool = False,
        use_markdown: bool = False,
        guidelines: Optional[str] = None,
        metadata_properties: List[AllowedMetadataPropertyTypes] = None,
        vectors_settings: List[VectorSettings] = None,
    ) -> "FeedbackDataset":
        """
        You can use this method to create a basic dataset for proximal policy optimization tasks.

        Args:
            rating_scale: Set this parameter to the number of relevancy scale you want to add to your dataset
            context: Set this parameter to True if you want to add context to your dataset
            use_markdown: Set this parameter to True if you want to use markdown in your dataset
            guidelines: Contains the guidelines for the dataset
            metadata_properties: Contains the metadata properties that will be indexed and could be used to filter the dataset. Defaults to `None`.
            vectors_settings: Define the configuration of the vectors associated to the records that will be used to perform the vector search. Defaults to `None`.

        Returns:
            A `FeedbackDataset` object for proximal policy optimization containing "context" and "action" fields and a LabelQuestion named "label"
        """
        default_guidelines = "This is a proximal policy optimization dataset that contains contexts and prompts. Please choose the label that best describes prompt."
        fields = [TextField(name="prompt", use_markdown=use_markdown)]
        if context:
            fields.append(TextField(name="context", use_markdown=use_markdown, required=False))

        return cls(
            fields=fields,
            questions=[
                RatingQuestion(
                    name="prompt",
                    values=list(range(1, rating_scale + 1)),
                    description="Choose one of the labels that best describes the prompt.",
                )
            ],
            guidelines=default_guidelines if guidelines is None else guidelines,
            metadata_properties=metadata_properties,
            vectors_settings=vectors_settings,
        )

    @classmethod
    def for_direct_preference_optimization(
        cls: Type["FeedbackDataset"],
        number_of_responses: int = 2,
        context: bool = False,
        use_markdown: bool = False,
        guidelines: Optional[str] = None,
        metadata_properties: List[AllowedMetadataPropertyTypes] = None,
        vectors_settings: List[VectorSettings] = None,
    ) -> "FeedbackDataset":
        """
        You can use this method to create a basic dataset for direct preference optimization tasks.

        Args:
            number_of_responses: Set this parameter to the number of responses you want to add to your dataset
            context: Set this parameter to True if you want to add context to your dataset
            use_markdown: Set this parameter to True if you want to use markdown in your dataset
            metadata_properties: Contains the metadata properties that will be indexed and could be used to filter the dataset. Defaults to `None`.
            vectors_settings: Define the configuration of the vectors associated to the records that will be used to perform the vector search. Defaults to `None`.

        Returns:
            A `FeedbackDataset` object for direct preference optimization containing "prompt", "response1", "response2" with the optional "context" fields and a RatingQuestion named "preference"
        """
        default_guidelines = "This is a direct preference optimization dataset that contains contexts and options. Please rank the options that you would prefer in the given context."
        return cls.for_preference_modeling(
            number_of_responses=number_of_responses,
            context=context,
            use_markdown=use_markdown,
            guidelines=default_guidelines if guidelines is None else guidelines,
            metadata_properties=metadata_properties,
            vectors_settings=vectors_settings,
        )

    @classmethod
    def for_retrieval_augmented_generation(
        cls: Type["FeedbackDataset"],
        number_of_retrievals: int = 1,
        rating_scale: int = 7,
        use_markdown: bool = False,
        guidelines: Optional[str] = None,
        metadata_properties: List[AllowedMetadataPropertyTypes] = None,
        vectors_settings: List[VectorSettings] = None,
    ) -> "FeedbackDataset":
        """
        You can use this method to create a basic dataset for retrieval augmented generation tasks.

        Args:
            number_of_retrievals: Set this parameter to the number of documents you want to add to your dataset
            rating_scale: Set this parameter to the number of relevancy scale you want to add to your dataset
            use_markdown: Set this parameter to True if you want to use markdown in your dataset
            guidelines: Contains the guidelines for the dataset
            metadata_properties: Contains the metadata properties that will be indexed and could be used to filter the dataset. Defaults to `None`.
            vectors_settings: Define the configuration of the vectors associated to the records that will be used to perform the vector search. Defaults to `None`.

        Returns:
            A `FeedbackDataset` object for retrieval augmented generation containing "query" and "retrieved_document" fields and a TextQuestion named "response"
        """
        default_guidelines = "This is a retrieval augmented generation dataset that contains queries and retrieved documents. Please rate the relevancy of retrieved document and write the response to the query in the response field."
        document_fields = [
            TextField(
                name="retrieved_document_" + str(doc + 1),
                title="Retrieved Document " + str(doc + 1),
                use_markdown=use_markdown,
                required=True if doc == 0 else False,
            )
            for doc in range(number_of_retrievals)
        ]

        rating_questions = [
            RatingQuestion(
                name="rating_retrieved_document_" + str(doc + 1),
                title="Rate the relevance of the Retrieved Document " + str(doc + 1) + " for the query",
                values=list(range(1, rating_scale + 1)),
                description="Rate the relevance of the retrieved document.",
                required=True if doc == 0 else False,
            )
            for doc in range(number_of_retrievals)
        ]

        total_questions = rating_questions + [
            TextQuestion(
                name="response",
                title="Write a helpful, harmless, accurate response to the query.",
                description="Write the response to the query.",
                use_markdown=use_markdown,
                required=False,
            )
        ]

        return cls(
            fields=[TextField(name="query", use_markdown=use_markdown, required=True)] + document_fields,
            questions=total_questions,
            guidelines=default_guidelines if guidelines is None else guidelines,
            metadata_properties=metadata_properties,
            vectors_settings=vectors_settings,
        )

    @classmethod
    def for_multi_modal_classification(
        cls: Type["FeedbackDataset"],
        labels: List[str],
        multi_label: bool = False,
        guidelines: Optional[str] = None,
        metadata_properties: List[AllowedMetadataPropertyTypes] = None,
        vectors_settings: List[VectorSettings] = None,
    ) -> "FeedbackDataset":
        """
        You can use this method to create a basic dataset for multi-modal (video, audio,image) classification tasks.

        Args:
            labels: A list of labels for your dataset
            multi_label: Set this parameter to True if you want to add multiple labels to your dataset
            guidelines: Contains the guidelines for the dataset
            metadata_properties: Contains the metadata properties that will be indexed and could be used to filter the dataset. Defaults to `None`.
            vectors_settings: Define the configuration of the vectors associated to the records that will be used to perform the vector search. Defaults to `None`.

        Returns:
            A `FeedbackDataset` object for multi-modal classification containing a "content" field with video, audio or image data and LabelQuestion or MultiLabelQuestion named "label"
        """
        default_guidelines = "This is a multi-modal classification dataset that contains videos, audios or images. Given a set of media files and a predefined set of labels, the goal of multi-modal classification is to assign one or more labels to each media file based on its content. Please classify the media file by making the correct selection."

        description = "Classify the media content by selecting the correct label from the given list of labels."
        return cls(
            fields=[TextField(name="content", use_markdown=True, required=True)],
            questions=[
                LabelQuestion(
                    name="label",
                    labels=labels,
                    description=description,
                )
                if not multi_label
                else MultiLabelQuestion(
                    name="label",
                    labels=labels,
                    description=description,
                )
            ],
            guidelines=guidelines
            if guidelines is not None
            else default_guidelines
            if multi_label
            else default_guidelines.replace("one or more labels", "one label"),
            metadata_properties=metadata_properties,
            vectors_settings=vectors_settings,
        )

    @classmethod
    def for_multi_modal_transcription(
        cls: Type["FeedbackDataset"],
        guidelines: Optional[str] = None,
        metadata_properties: List[AllowedMetadataPropertyTypes] = None,
        vectors_settings: List[VectorSettings] = None,
    ) -> "FeedbackDataset":
        """
        You can use this method to create a basic dataset for multi-modal (video, audio,image) transcription tasks.

        Args:
            use_markdown: Set this parameter to True if you want to use markdown in your TextQuestion. Defaults to `False`.
            guidelines: Contains the guidelines for the dataset
            metadata_properties: Contains the metadata properties that will be indexed and could be used to filter the dataset. Defaults to `None`.
            vectors_settings: Define the configuration of the vectors associated to the records that will be used to perform the vector search. Defaults to `None`.

        Returns:
            A `FeedbackDataset` object for multi-modal transcription containing a "content" field with video, audio or image data and a TextQuestion named "description"
        """
        default_guidelines = "This is a multi-modal transcription dataset that contains video, audio or image data. Please describe the media content."
        return cls(
            fields=[
                TextField(name="content", use_markdown=True, required=True),
            ],
            questions=[
                TextQuestion(
                    name="description",
                    description="Provide a description of the media content, detailing the specific characteristics or elements present in each file.",
                    use_markdown=True,
                    required=True,
                )
            ],
            guidelines=default_guidelines if guidelines is None else guidelines,
            metadata_properties=metadata_properties,
            vectors_settings=vectors_settings,
        )



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Union
from uuid import UUID

from argilla.pydantic_v1 import BaseModel, Field, StrictInt, StrictStr, conint, root_validator


class FeedbackDatasetModel(BaseModel):
    id: UUID
    name: str = Field(regex="^(?!-|_)[a-zA-Z0-9-_ ]+$")
    guidelines: Optional[str] = None
    # Set default for backward compatibility
    allow_extra_metadata: Optional[bool] = True
    status: Optional[str] = None
    workspace_id: Optional[UUID] = None
    last_activity_at: Optional[datetime] = None
    inserted_at: datetime
    updated_at: datetime


class FeedbackRankingValueModel(BaseModel):
    value: StrictStr
    rank: Optional[conint(ge=1)] = None


class FeedbackValueModel(BaseModel):
    value: Union[StrictStr, StrictInt, List[str], List[FeedbackRankingValueModel]]


class FeedbackResponseStatus(str, Enum):
    draft = "draft"
    submitted = "submitted"
    discarded = "discarded"


class FeedbackResponseStatusFilter(str, Enum):
    draft = "draft"
    missing = "missing"  # not a status as-is, used as a filter to indicate no response
    submitted = "submitted"
    discarded = "discarded"


# TODO: these models shouldn't transform the payload from the server to not JSON serializable types (UUID, datetime, etc)


class FeedbackResponseModel(BaseModel):
    id: UUID
    values: Union[Dict[str, FeedbackValueModel], None]
    status: FeedbackResponseStatus
    user_id: UUID
    inserted_at: datetime
    updated_at: datetime


class FeedbackSuggestionModel(BaseModel):
    id: UUID
    question_id: str
    type: Optional[Literal["human", "model"]] = None
    score: Optional[float] = None
    value: Any
    agent: Optional[str] = None


class FeedbackItemModel(BaseModel):
    id: UUID
    fields: Dict[str, Any]
    metadata: Optional[Dict[str, Any]] = None
    external_id: Optional[str] = None
    responses: Optional[List[FeedbackResponseModel]] = []
    vectors: Optional[Dict[str, List[float]]] = None
    suggestions: Optional[List[FeedbackSuggestionModel]] = []
    vectors: Optional[Dict[str, List[float]]] = {}
    inserted_at: datetime
    updated_at: datetime


class FeedbackRecordsModel(BaseModel):
    items: List[FeedbackItemModel]
    total: int


# TODO: `query_score` naming can be improved to simply `score`. (frontend should be aligned)
# TODO: Maybe `query_score` should not be optional.
class FeedbackRecordSearchModel(BaseModel):
    record: FeedbackItemModel
    query_score: Optional[float]


class FeedbackRecordsSearchModel(BaseModel):
    items: List[FeedbackRecordSearchModel]
    total: int


class FeedbackRecordsSearchVectorQuery(BaseModel):
    name: str
    record_id: Optional[UUID] = None
    value: Optional[List[float]] = None

    @root_validator(skip_on_failure=True)
    def check_required(cls, values: dict) -> dict:
        """Check that either 'record_id' or 'value' is provided"""
        record_id = values.get("record_id")
        value = values.get("value")

        if bool(record_id) == bool(value):
            raise ValueError("Either 'record_id' or 'value' must be provided")

        return values


class FeedbackFieldModel(BaseModel):
    id: UUID
    name: str
    title: str
    required: bool
    settings: Dict[str, Any]
    inserted_at: datetime
    updated_at: datetime


class FeedbackQuestionModel(BaseModel):
    id: UUID
    name: str
    title: str
    description: Optional[str] = None
    required: bool
    settings: Dict[str, Any]
    inserted_at: datetime
    updated_at: datetime


class FeedbackMetadataPropertyModel(BaseModel):
    id: UUID
    name: str
    title: str
    visible_for_annotators: bool
    settings: Dict[str, Any]
    inserted_at: datetime
    updated_at: datetime


class FeedbackRecordsMetricsModel(BaseModel):
    count: int


class FeedbackVectorSettingsModel(BaseModel):
    id: UUID
    name: str
    title: str
    dimensions: int
    inserted_at: datetime
    updated_at: datetime


class FeedbackListVectorSettingsModel(BaseModel):
    items: List[FeedbackVectorSettingsModel]


class FeedbackMetricsModel(BaseModel):
    records: FeedbackRecordsMetricsModel



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from argilla.client.feedback.integrations.huggingface.model_card.model_card import (
    ArgillaModelCard,
    FrameworkCardData,
    OpenAIModelCardData,
    PeftModelCardData,
    SentenceTransformerCardData,
    SetFitModelCardData,
    SpacyModelCardData,
    SpacyTransformersModelCardData,
    TransformersModelCardData,
    TRLModelCardData,
)

__all__ = [
    "ArgillaModelCard",
    "FrameworkCardData",
    "SentenceTransformerCardData",
    "TransformersModelCardData",
    "SetFitModelCardData",
    "PeftModelCardData",
    "SpacyModelCardData",
    "SpacyTransformersModelCardData",
    "OpenAIModelCardData",
    "TRLModelCardData",
]



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import warnings
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union
from uuid import UUID

from argilla.client.feedback.schemas.enums import ResponseStatus
from argilla.client.feedback.schemas.records import FeedbackRecord, ResponseSchema, SuggestionSchema
from argilla.client.feedback.schemas.remote.shared import RemoteSchema
from argilla.client.sdk.users.models import UserRole
from argilla.client.sdk.v1.records import api as records_api_v1
from argilla.client.sdk.v1.suggestions import api as suggestions_api_v1
from argilla.client.utils import allowed_for_roles
from argilla.pydantic_v1 import Field

if TYPE_CHECKING:
    import httpx

    from argilla.client.sdk.v1.datasets.models import FeedbackResponseModel, FeedbackSuggestionModel
    from argilla.client.sdk.v1.records.models import FeedbackRecordModel


class RemoteSuggestionSchema(SuggestionSchema, RemoteSchema):
    question_id: UUID

    class Config:
        validate_assignment = True
        allow_mutation = False

    @allowed_for_roles(roles=[UserRole.owner, UserRole.admin])
    def delete(self) -> None:
        """Deletes the `RemoteSuggestionSchema` from Argilla."""
        try:
            suggestions_api_v1.delete_suggestion(client=self.client, id=self.id)
        except Exception as e:
            raise RuntimeError(f"Failed to delete suggestion with ID `{self.id}` from Argilla.") from e

    def to_local(self) -> "SuggestionSchema":
        """Converts the `RemoteSuggestionSchema` to a `SuggestionSchema`."""
        return SuggestionSchema(
            question_name=self.question_name,
            type=self.type,
            score=self.score,
            value=self.value,
            agent=self.agent,
        )

    @classmethod
    def from_api(
        cls,
        payload: "FeedbackSuggestionModel",
        question_id_to_name: Dict[UUID, str],
        client: Optional["httpx.Client"] = None,
    ) -> "RemoteSuggestionSchema":
        return RemoteSuggestionSchema(
            id=payload.id,
            client=client,
            question_id=payload.question_id,
            question_name=question_id_to_name[UUID(payload.question_id)],
            type=payload.type,
            score=payload.score,
            value=payload.value,
            agent=payload.agent,
        )


class RemoteResponseSchema(ResponseSchema, RemoteSchema):
    inserted_at: datetime
    updated_at: datetime

    def to_local(self) -> "ResponseSchema":
        """Converts the `RemoteResponseSchema` to a `ResponseSchema`."""
        return ResponseSchema(
            user_id=self.user_id,
            values=self.values,
            status=self.status,
        )

    @classmethod
    def from_api(cls, payload: "FeedbackResponseModel") -> "RemoteResponseSchema":
        return RemoteResponseSchema(
            user_id=payload.user_id,
            values=payload.values,
            status=ResponseStatus(payload.status),
            inserted_at=payload.inserted_at,
            updated_at=payload.updated_at,
        )


AllowedSuggestionSchema = Union[RemoteSuggestionSchema, SuggestionSchema]


class RemoteFeedbackRecord(FeedbackRecord, RemoteSchema):
    """Schema for the records of a `RemoteFeedbackDataset`.

    Note this schema shouldn't be instantiated directly, but just internally by the
    `RemoteFeedbackDataset` class when fetching records from Argilla.

    Args:
        question_name_to_id: A dictionary that maps the question names to their corresponding IDs.
        responses: A list of `RemoteResponseSchema` that contains the responses for the
            current record in Argilla. Every response is linked to only one user. Defaults
            to an empty list.
        suggestions: A list of `RemoteSuggestionSchema` that contains the suggestions
            for the current record in Argilla. Every suggestion is linked to only one
            question. Defaults to an empty list.
    """

    # TODO: remote record should receive a dataset instead of this
    question_name_to_id: Optional[Dict[str, UUID]] = Field(..., exclude=True, repr=False)

    responses: List[RemoteResponseSchema] = Field(default_factory=list)
    suggestions: Union[Tuple[AllowedSuggestionSchema], List[AllowedSuggestionSchema]] = Field(default_factory=tuple)

    class Config:
        allow_mutation = True
        validate_assignment = True

    def __normalize_suggestions_to_update(
        self,
        suggestions: Union[
            Dict[str, Any], List[Dict[str, Any]], AllowedSuggestionSchema, List[AllowedSuggestionSchema]
        ],
    ) -> List[AllowedSuggestionSchema]:
        if isinstance(suggestions, (dict, SuggestionSchema)):
            suggestions = [suggestions]

        new_suggestions = {}

        for suggestion in suggestions:
            if isinstance(suggestion, dict):
                if "id" in suggestion:
                    if "question_id" not in suggestion or not suggestion["question_id"]:
                        suggestion["question_id"] = self.question_name_to_id[suggestion["question_name"]]
                    suggestion = RemoteSuggestionSchema(client=self.client, **suggestion)
                else:
                    suggestion = SuggestionSchema(**suggestion)

            if suggestion.question_name in new_suggestions:
                warnings.warn(
                    f"A suggestion for question `{suggestion.question_name}` has been"
                    " provided twice in the same update, so the last one will be the one"
                    " to be kept.",
                    UserWarning,
                    stacklevel=1,
                )

            new_suggestions[suggestion.question_name] = suggestion

        return list(new_suggestions.values())

    @allowed_for_roles(roles=[UserRole.owner, UserRole.admin])
    def update(
        self,
        suggestions: Optional[
            Union[
                SuggestionSchema,
                Dict[str, Any],
                List[SuggestionSchema],
                List[Dict[str, Any]],
            ]
        ] = None,
    ) -> None:
        """Update a `RemoteFeedbackRecord` in Argilla.

        Note that this method will update the record in Argilla directly.

        Args:
            suggestions: **DEPRECATED** the new list of suggestions to set for the
                record. It can be a single `SuggestionSchema`, a list of `SuggestionSchema`,
                a single dictionary or a list of dictionary. If a dictionary is provided,
                it will be converted to a `SuggestionSchema` internally. Defaults to
                `None`.

        Raises:
            PermissionError: if the user does not have either `owner` or `admin` role.
        """
        if suggestions is not None:
            warnings.warn(
                "The `suggestions` argument is deprecated and will be removed in "
                "future releases. In order to update the suggestions of a record "
                "assign the `suggestions` attribute directly.",
                DeprecationWarning,
                stacklevel=1,
            )

            new_suggestions = self.__normalize_suggestions_to_update(suggestions)
            self.suggestions = tuple(new_suggestions)

        self.__updated_record_data()

    def __updated_record_data(self) -> None:
        response = records_api_v1.update_record(self.client, self.id, self.to_server_payload(self.question_name_to_id))

        updated_record = self.from_api(
            payload=response.parsed,
            question_id_to_name={value: key for key, value in self.question_name_to_id.items()}
            if self.question_name_to_id
            else None,
            client=self.client,
        )

        self.__dict__.update(updated_record.__dict__)

    @allowed_for_roles(roles=[UserRole.owner, UserRole.admin])
    def delete_suggestions(self, suggestions: Union[RemoteSuggestionSchema, List[RemoteSuggestionSchema]]) -> None:
        """Deletes the provided suggestions from the record in Argilla. Note that the
        suggestions must exist in Argilla to be removed from the record.

        Args:
            suggestions: can be a single `RemoteSuggestionSchema` or a list of
                `RemoteSuggestionSchema`.

        Raises:
            PermissionError: if the user does not have either `owner` or `admin` role.

        **DEPRECATED**
        """

        warnings.warn(
            "The `delete_suggestions` method is deprecated and will be removed in "
            "future releases. In order to update/delete the suggestions of a record "
            "assign the `suggestions` attribute directly (`record.suggestions = []`).",
            DeprecationWarning,
            stacklevel=1,
        )

        if isinstance(suggestions, RemoteSuggestionSchema):
            suggestions = [suggestions]

        existing_suggestions = {suggestion.question_name: suggestion for suggestion in self.suggestions}
        delete_suggestions = []
        for suggestion in suggestions:
            if suggestion.question_name not in existing_suggestions:
                warnings.warn(
                    f"A suggestion for question `{suggestion.question_name}` has not been"
                    " provided, so it cannot be removed.",
                    UserWarning,
                    stacklevel=1,
                )
            else:
                existing_suggestions.pop(suggestion.question_name, None)
                delete_suggestions.append(suggestion)

        try:
            records_api_v1.delete_suggestions(
                client=self.client, id=self.id, suggestion_ids=[suggestion.id for suggestion in delete_suggestions]
            )
            self.__dict__["suggestions"] = tuple(existing_suggestions.values())
        except Exception as e:
            raise RuntimeError(
                f"Failed to delete suggestions with IDs `{[suggestion.id for suggestion in delete_suggestions]}` from "
                f"record with ID `{self.id}` from Argilla."
            ) from e

    @allowed_for_roles(roles=[UserRole.owner, UserRole.admin])
    def delete(self) -> FeedbackRecord:
        """Deletes the `RemoteFeedbackRecord` from Argilla.

        Returns:
            The deleted record formatted as a `FeedbackRecord`.

        Raises:
            PermissionError: if the user does not have either `owner` or `admin` role.
        """
        try:
            response = records_api_v1.delete_record(client=self.client, id=self.id)
        except Exception as e:
            raise RuntimeError(f"Failed to delete record with ID `{self.id}` from Argilla.") from e
        return RemoteFeedbackRecord.from_api(
            payload=response.parsed, question_id_to_name={value: key for key, value in self.question_name_to_id.items()}
        ).to_local()

    def to_local(self) -> "FeedbackRecord":
        """Converts the `RemoteFeedbackRecord` to a `FeedbackRecord`."""
        return FeedbackRecord(
            fields=self.fields,
            responses=[response.to_local() for response in self.responses],
            suggestions=[suggestion.to_local() for suggestion in self.suggestions],
            metadata=self.metadata,
            vectors=self.vectors,
            external_id=self.external_id,
        )

    @classmethod
    def from_api(
        cls,
        payload: "FeedbackRecordModel",
        question_id_to_name: Dict[UUID, str],
        client: Optional["httpx.Client"] = None,
    ) -> "RemoteFeedbackRecord":
        return RemoteFeedbackRecord(
            id=payload.id,
            client=client,
            fields=payload.fields,
            responses=[RemoteResponseSchema.from_api(response) for response in payload.responses]
            if payload.responses
            else [],
            suggestions=[
                RemoteSuggestionSchema.from_api(suggestion, question_id_to_name=question_id_to_name, client=client)
                for suggestion in payload.suggestions
            ]
            if payload.suggestions
            else [],
            metadata=payload.metadata if payload.metadata else {},
            vectors=payload.vectors if payload.vectors else {},
            external_id=payload.external_id if payload.external_id else None,
            question_name_to_id={value: key for key, value in question_id_to_name.items()},
        )
